"""Type conversion utilities"""
from collections import UserList
from datetime import date, datetime, timedelta
from dateutil.parser import UnknownTimezoneWarning  # type: ignore  # (missing from type stubs)
from dateutil.parser import parse as parse_date
from dateutil.tz import tzlocal, tzoffset
from io import BytesIO
from logging import getLogger
from os.path import abspath, expanduser
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Union
from warnings import catch_warnings, simplefilter

from pyinaturalist.constants import (
    AnyFile,
    Coordinates,
    Dimensions,
    HistogramResponse,
    JsonResponse,
    ResponseResult,
)

GENERIC_TIME_FIELDS = ('created_at', 'last_post_at', 'updated_at')
OBSERVATION_TIME_FIELDS = (
    'created_at_details',
    'created_time_zone',
    'observed_on',
    'observed_on_details',
    'observed_on_string',
    'observed_time_zone',
    'time_zone_offset',
)
logger = getLogger(__name__)


# Wrapper functions to apply conversions to all results in a response
# --------------------


def convert_all_coordinates(results: List[ResponseResult]) -> List[ResponseResult]:
    """Convert coordinate pairs in response items from strings to floats, if valid

    Args:
        results: Results from API response; expects coordinates in either 'location' key or
            'latitude' and 'longitude' keys
    """
    results = [convert_lat_long_dict(result) for result in results]
    results = [convert_lat_long_list(result) for result in results]
    return results


def convert_all_place_coordinates(response: JsonResponse) -> JsonResponse:
    """Convert locations for both standard and community-contributed places to floats"""
    response['results'] = {
        'standard': convert_all_coordinates(response['results'].get('standard')),
        'community': convert_all_coordinates(response['results'].get('community')),
    }
    return response


def convert_all_timestamps(results: List[ResponseResult]) -> List[ResponseResult]:
    """Replace all date/time info with datetime objects, where possible"""
    results = [convert_generic_timestamps(result) for result in results]
    results = [convert_observation_timestamps(result) for result in results]
    return results


def convert_histogram(response: JsonResponse) -> HistogramResponse:
    """Convert a response containing time series data into a single ``{date: value}`` dict"""
    # The inner result object's key will be the name of the interval requested
    interval = next(iter(response['results'].keys()))
    histogram = response['results'][interval]

    # Convert keys to appropriate type depending on interval
    if interval in ['month_of_year', 'week_of_year']:
        return {int(k): v for k, v in histogram.items()}
    else:
        return {parse_date(k): v for k, v in histogram.items()}


# Type conversion functions for individual results or values
# --------------------


def convert_csv_list(obj: Any) -> str:
    """Convert list parameters into an API-compatible (comma-delimited) string"""
    if not obj:
        return ''
    elif isinstance(obj, list):
        return ','.join(map(str, obj))
    else:
        return str(obj)


def convert_isoformat(value: Union[date, datetime, str]) -> str:
    """Convert a date, datetime, or timestamps to a string in ISO 8601 format.
    If it's a datetime and doesn't already have tzinfo, set it to the system's local timezone.

    Raises:
        :py:exc:`dateutil.parser._parser.ParserError` if a date/datetime format is invalid
    """
    if isinstance(value, str):
        value = parse_date(value)
    if isinstance(value, datetime) and not value.tzinfo:
        value = value.replace(tzinfo=tzlocal())
    return value.isoformat()


def convert_lat_long(obj: Union[Dict, List, None, str]) -> Optional[Coordinates]:
    """Convert a coordinate pair as a dict, list, or string into a pair of floats, if valid"""
    if not obj:
        return None
    if isinstance(obj, str):
        return try_float_pair(*str(obj).split(','))
    elif isinstance(obj, list):
        return try_float_pair(*obj)
    elif isinstance(obj, dict):
        return try_float_pair(obj.get('latitude'), obj.get('longitude'))


def convert_lat_long_dict(result: ResponseResult) -> ResponseResult:
    """Convert a coordinate pair dict within a response to floats, if valid"""
    if 'latitude' in result and 'longitude' in result:
        result['latitude'] = try_float(result['latitude'])
        result['longitude'] = try_float(result['longitude'])
    return result


def convert_lat_long_list(result: ResponseResult):
    """Convert a coordinate pairs in a response item from strings to floats, if valid"""
    # Format inner record if present, e.g. for search results
    if 'record' in result:
        result['record'] = convert_lat_long_list(result['record'])
        return result

    if ',' in (result.get('location') or ''):
        result['location'] = [try_float(coord) for coord in result['location'].split(',')]
    return result


def convert_generic_timestamps(result: ResponseResult) -> ResponseResult:
    """Replace generic created/updated info that's returned by multiple endpoints.
    **Note:** Compared to observation timestamps, these are generally more reliable. These seem to
    be consistently in ISO 8601 format.
    """
    # Format inner record if present, e.g. for search results
    if 'record' in result:
        result['record'] = convert_generic_timestamps(result['record'])
        return result

    for field in GENERIC_TIME_FIELDS:
        datetime_obj = try_datetime(result.get(field, ''))
        if datetime_obj:
            result[field] = datetime_obj
    return result


# TODO: Replace this with attrs version
def convert_observation_timestamps(result: ResponseResult) -> ResponseResult:
    """Replace observation date/time info with datetime objects"""
    if 'created_at_details' not in result and 'observed_on_string' not in result:
        return result

    observation = result.copy()
    tz_offset = observation.get('time_zone_offset', '')
    tz_name = observation.get('observed_time_zone')

    created_datetime = observation.get('created_at')
    if not isinstance(created_datetime, datetime):
        created_datetime = try_datetime(observation.get('created_at_details', {}).get('date'))
        created_datetime = convert_offset(created_datetime, tz_offset, tz_name)

    # Ignore any timezone info in observed_on timestamp; offset field is more reliable
    observed_datetime = try_datetime(observation.get('observed_on_string', ''), ignoretz=True)
    observed_datetime = convert_offset(observed_datetime, tz_offset, tz_name)

    # If valid, add the datetime objects and remove all other redundant date/time fields
    if created_datetime and observed_datetime:
        for field in OBSERVATION_TIME_FIELDS:
            observation.pop(field, None)
    observation['created_at'] = created_datetime
    observation['observed_on'] = observed_datetime

    return observation


def convert_observation_timestamp(
    timestamp: str, tz_offset: str = None, tz_name: str = None, ignoretz: bool = False
) -> Optional[datetime]:
    """Convert an observation timestamp + timezone info to a datetime. This is needed because
    observed_on and created_at can be in in inconsistent (user-submitted?) formats.
    """
    dt = try_datetime(timestamp, ignoretz=ignoretz)
    return convert_offset(dt, tz_offset, tz_name)


def convert_offset(
    datetime_obj: Optional[datetime], tz_offset: str = None, tz_name: str = None
) -> Optional[datetime]:
    """Use timezone offset info to replace a datetime's tzinfo"""
    if not datetime_obj or not tz_offset:
        return datetime_obj

    try:
        offset = parse_offset(tz_offset, tz_name)
        return datetime_obj.replace(tzinfo=offset)
    except (AttributeError, TypeError, ValueError) as e:
        logger.debug(f'Could not parse offset: {tz_offset}: {str(e)}')
        return None


def parse_offset(tz_offset: str, tz_name: str = None) -> tzoffset:
    """Convert a timezone offset string to a tzoffset object, accounting for some common variations
    in format

    Examples:

        >>> parse_offset('GMT-08:00', 'PST')
        tzoffset('PST', -28800)
        >>> parse_offset('-06:00')
        tzoffset(None, -21600)
        >>> parse_offset('+05:30')
        tzoffset(None, 19800)
        >>> parse_offset('0530')
        tzoffset(None, 19800)

    """

    def remove_prefixes(text, prefixes):
        for prefix in prefixes:
            if text.startswith(prefix):
                text = text[len(prefix) :]  # noqa  # black and flake8 fight over this one
        return text

    # Parse hours, minutes, and sign from offset string; account for either `hh:mm` or `hhmm` format
    tz_offset = remove_prefixes(tz_offset, ['GMT', 'UTC']).strip()
    multiplier = -1 if tz_offset.startswith('-') else 1
    tz_offset = tz_offset.lstrip('+-')
    if ':' in tz_offset:
        hours, minutes = tz_offset.split(':')
    else:
        hours, minutes = tz_offset[:2], tz_offset[2:]

    # Convert to a timezone offset in +/- seconds
    delta = timedelta(hours=int(hours), minutes=int(minutes))
    return tzoffset(tz_name, delta.total_seconds() * multiplier)


def safe_split(value: Any, delimiter: str = '|') -> List:
    """Split a pipe-(or other token)-delimited string"""
    return ensure_list(value, convert_csv=True, delimiter=delimiter)


def strip_empty_values(values: Dict) -> Dict:
    """Remove any dict items with empty or ``None`` values."""
    return {k: v for k, v in values.items() if v or v in [False, 0, 0.0]}


# ?
# -------------------


def ensure_list(value: Any, convert_csv: bool = False, delimiter: str = ',') -> List[Any]:
    """Convert an object, response, or (optionally) comma-separated string into a list"""
    if not value:
        return []
    elif isinstance(value, dict) and 'results' in value:
        value = value['results']
    elif convert_csv and isinstance(value, str) and delimiter in value:
        return [s.strip() for s in value.split(delimiter)]

    if isinstance(value, (UserList, list, tuple, set)):
        return list(value)
    else:
        return [value]


def ensure_file_obj(value: AnyFile) -> IO:
    """Given a file object or path, read it into a file-like object if it's a path"""
    if isinstance(value, (str, Path)):
        file_path = abspath(expanduser(value))
        logger.info(f'Reading from file: {file_path}')
        with open(file_path, 'rb') as f:
            return BytesIO(f.read())
    return value


# Formatting functions
# --------------------


def format_dimensions(dimensions: Union[Dict[str, int], Dimensions, None]) -> Dimensions:
    """Simplify a 'dimensions' dict into a ``(width, height)`` tuple"""
    if not dimensions:
        return (0, 0)
    if isinstance(dimensions, tuple):
        return dimensions
    return dimensions.get("width", 0), dimensions.get("height", 0)


def format_file_size(n_bytes: int) -> str:
    """Convert a file size in bytes into a human-readable format"""
    filesize = float(n_bytes or 0)

    def _format(unit):
        return f'{int(filesize)} {unit}' if unit == 'bytes' else f'{filesize:.2f} {unit}'

    for unit in ['bytes', 'KB', 'MB', 'GB']:
        if filesize < 1024 or unit == 'GB':
            return _format(unit)
        filesize /= 1024

    return _format(unit)


def format_license(value: str) -> str:
    """Format a Creative Commons license code"""
    return str(value).upper().replace('_', '-')


# 'Safe' conversion functions that return invalid values as None instead of raising an error
# --------------------


def try_datetime(timestamp: Any, **kwargs) -> Optional[datetime]:
    """Parse a timestamp string into a datetime, if valid; return ``None`` otherwise"""
    if isinstance(timestamp, datetime):
        return timestamp
    if not timestamp or not str(timestamp).strip():
        return None

    try:
        # Suppress UnknownTimezoneWarning
        with catch_warnings():
            simplefilter('ignore', category=UnknownTimezoneWarning)
            return parse_date(timestamp, **kwargs)
    except (AttributeError, TypeError, ValueError) as e:
        logger.debug(f'Could not parse timestamp: {timestamp}: "{str(e)}"')
        return None


def try_float(value: Any) -> Optional[float]:
    """Convert a value to a float, if valid; return ``None`` otherwise"""
    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def try_float_pair(*values: Any) -> Optional[Coordinates]:
    """Convert a pair of coordinat values to floats, if both are valid; return ``None`` otherwise"""
    if len(values) != 2:
        return None
    try:
        return float(values[0]), float(values[1])
    except (TypeError, ValueError):
        return None


def try_int(value: Any) -> Optional[float]:
    """Convert a value to a int, if valid; return ``None`` otherwise"""
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def try_int_or_float(value: Any) -> Union[int, float, None]:
    """Convert a value to either an int or a float, if valid; return ``None`` otherwise"""
    return try_int(str(value)) or try_float(str(value))



from logging import getLogger

from pyinaturalist.api_docs import document_request_params
from pyinaturalist.api_docs import templates as docs
from pyinaturalist.constants import JsonResponse
from pyinaturalist.converters import convert_all_timestamps, convert_generic_timestamps
from pyinaturalist.v1 import get_v1

logger = getLogger(__name__)


def get_user_by_id(user_id: int, **params) -> JsonResponse:
    """Get a user by ID.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Users/get_users_id

    Args:
        user_id: Get the user with this ID. Only a single ID is allowed per request.

    Example:

        >>> response = get_user_by_id(123456)
        >>> pprint(response)
        [1234] my_username

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_user_by_id.py

    Returns:
        Response dict containing user record
    """
    response = get_v1('users', ids=[user_id], **params)
    results = response.json()['results']
    if not results:
        return {}
    return convert_generic_timestamps(results[0])


@document_request_params([docs._search_query, docs._project_id, docs._pagination])
def get_users_autocomplete(q: str, **params) -> JsonResponse:
    """Given a query string, return users with names or logins starting with the search term

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Users/get_users_autocomplete

    Note: Pagination is supported; default page size is 6, and max is 100.

    Example:

        >>> response = get_taxa_autocomplete(q='my_userna')
        >>> pprint(response)
        [1234] my_username
        [12345] my_username_2

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_users_autocomplete.py

    Returns:
        Response dict containing user records
    """
    response = get_v1('users/autocomplete', q=q, **params)
    users = response.json()
    users['results'] = convert_all_timestamps(users['results'])
    return users



from datetime import datetime
from typing import Dict, List, Optional

from attr import define, field

from pyinaturalist.constants import Coordinates, TableRow
from pyinaturalist.converters import convert_observation_timestamp
from pyinaturalist.models import (
    Annotation,
    BaseModel,
    Comment,
    Identification,
    LazyProperty,
    ObservationFieldValue,
    Photo,
    ProjectObservation,
    Taxon,
    User,
    add_lazy_attrs,
    coordinate_pair,
    datetime_attr,
    datetime_now_attr,
    kwarg,
)
from pyinaturalist.v1 import get_observation


@define(auto_attribs=False, init=False, field_transformer=add_lazy_attrs)
class Observation(BaseModel):
    """An observation, based the schema of
    `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    created_at: datetime = datetime_now_attr
    cached_votes_total: int = kwarg
    captive: bool = kwarg
    comments_count: int = kwarg
    community_taxon_id: int = kwarg
    description: str = kwarg
    faves_count: int = kwarg
    geoprivacy: str = kwarg  # Enum
    id: int = kwarg
    id_please: bool = kwarg
    identifications_count: int = kwarg
    identifications_most_agree: bool = kwarg
    identifications_most_disagree: bool = kwarg
    identifications_some_agree: bool = kwarg
    license_code: str = kwarg  # Enum
    location: Optional[Coordinates] = coordinate_pair
    map_scale: int = kwarg
    mappable: bool = kwarg
    num_identification_agreements: int = kwarg
    num_identification_disagreements: int = kwarg
    oauth_application_id: str = kwarg
    obscured: bool = kwarg
    out_of_range: bool = kwarg
    owners_identification_from_vision: bool = kwarg
    place_guess: str = kwarg
    positional_accuracy: int = kwarg
    public_positional_accuracy: int = kwarg
    quality_grade: str = kwarg  # Enum
    site_id: int = kwarg
    species_guess: str = kwarg
    observed_on: Optional[datetime] = datetime_attr
    updated_at: Optional[datetime] = datetime_attr
    uri: str = kwarg
    uuid: str = kwarg

    # Nested collections
    faves: List = field(factory=list)
    outlinks: List = field(factory=list)
    place_ids: List = field(factory=list)
    preferences: Dict = field(factory=dict)
    project_ids: List = field(factory=list)
    project_ids_with_curator_id: List = field(factory=list)
    project_ids_without_curator_id: List = field(factory=list)
    quality_metrics: List = field(factory=list)
    reviewed_by: List = field(factory=list)
    sounds: List = field(factory=list)
    tags: List = field(factory=list)
    votes: List = field(factory=list)

    # Lazy-loaded nested model objects
    annotations: property = LazyProperty(Annotation.from_json_list)
    comments: property = LazyProperty(Comment.from_json_list)
    identifications: property = LazyProperty(Identification.from_json_list)
    ofvs: property = LazyProperty(ObservationFieldValue.from_json_list)
    photos: property = LazyProperty(Photo.from_json_list)
    project_observations: property = LazyProperty(ProjectObservation.from_json_list)
    taxon: property = LazyProperty(Taxon.from_json)
    user: property = LazyProperty(User.from_json)

    # Additional attributes from API response that aren't needed; just left here for reference
    # created_at_details: Dict = field(factory=dict)
    # created_time_zone: str = kwarg
    # flags: List = field(factory=list)
    # geojson: Dict = field(factory=dict)
    # non_owner_ids: List = field(factory=list)
    # observed_on_details: Dict = field(factory=dict)
    # observed_on_string: str = kwarg
    # observation_photos: List[Photo] = field(converter=Photo.from_dict_list, factory=list)
    # observed_time_zone: str = kwarg
    # spam: bool = kwarg
    # time_observed_at: Optional[datetime] = datetime_attr
    # time_zone_offset: str = kwarg

    # Attributes that will only be used during init and then omitted
    temp_attrs = [
        'created_at_details',
        'observed_on_string',
        'observed_time_zone',
        'time_zone_offset',
    ]

    # Convert observation timestamps prior to attrs init
    # TODO: better function signature for docs; use forge?
    def __init__(
        self,
        created_at_details: Dict = None,
        observed_on_string: str = None,
        observed_time_zone: str = None,
        time_zone_offset: str = None,
        **kwargs,
    ):
        tz_offset = time_zone_offset
        tz_name = observed_time_zone
        created_date = (created_at_details or {}).get('date')

        if not isinstance(kwargs.get('created_at'), datetime) and created_date:
            kwargs['created_at'] = convert_observation_timestamp(created_date, tz_offset, tz_name)
        if not isinstance(kwargs.get('observed_on'), datetime) and observed_on_string:
            kwargs['observed_on'] = convert_observation_timestamp(
                observed_on_string, tz_offset, tz_name, ignoretz=True
            )

        self.__attrs_init__(**kwargs)  # type: ignore

    @classmethod
    def from_id(cls, id: int):
        """Lookup and create a new Observation object from an ID"""
        json = get_observation(id)
        return cls.from_json(json)

    @property
    def thumbnail_url(self) -> Optional[str]:
        if not self.photos:
            return None
        return self.photos[0].thumbnail_url

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Taxon ID': self.taxon.id,
            'Taxon': self.taxon.full_name,
            'Observed on': self.observed_on,
            'User': self.user.login,
            'Location': self.place_guess or self.location,
        }

    def __str__(self) -> str:
        return (
            f'[{self.id}] {self.taxon.full_name} '
            f'observed on {self.observed_on} by {self.user.login} '
            f'at {self.place_guess or self.location}'
        )



from datetime import datetime
from dateutil.tz import tzutc

from pyinaturalist.constants import API_V1_BASE_URL
from pyinaturalist.v1 import get_user_by_id, get_users_autocomplete
from test.conftest import load_sample_data


def test_get_user_by_id(requests_mock):
    user_id = 1
    requests_mock.get(
        f'{API_V1_BASE_URL}/users/{user_id}',
        json=load_sample_data('get_user_by_id.json'),
        status_code=200,
    )

    user = get_user_by_id(user_id)
    assert user['id'] == user_id
    assert user['created_at'] == datetime(2008, 3, 20, 21, 15, 42, tzinfo=tzutc())
    assert user['spam'] is False


def test_get_users_autocomplete(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/users/autocomplete',
        json=load_sample_data('get_users_autocomplete.json'),
        status_code=200,
    )

    response = get_users_autocomplete(q='niconoe')
    first_result = response['results'][0]

    assert len(response['results']) == response['total_results'] == 3
    assert first_result['id'] == 886482
    assert first_result['login'] == 'niconoe'
    assert first_result['created_at'] == datetime(2018, 4, 23, 17, 11, 14, tzinfo=tzutc())



from typing import Any

from pyinaturalist.api_docs import document_request_params
from pyinaturalist.api_docs import templates as docs
from pyinaturalist.api_requests import get, put
from pyinaturalist.constants import API_V0_BASE_URL, JsonResponse
from pyinaturalist.converters import convert_all_timestamps
from pyinaturalist.pagination import add_paginate_all


@document_request_params([docs._search_query, docs._pagination])
@add_paginate_all(method='page')
def get_observation_fields(**params) -> JsonResponse:
    """Search observation fields. Observation fields are basically typed data fields that
    users can attach to observation.

    **API reference:** https://www.inaturalist.org/pages/api+reference#get-observation_fields

    Example:

        >>> get_observation_fields(q='number of individuals')
        >>> # Show just observation field IDs and names
        >>> from pprint import pprint
        >>> pprint({r['id']: r['name'] for r in response})

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_fields.py

    Returns:
        Observation fields as a list of dicts
    """
    response = get(f'{API_V0_BASE_URL}/observation_fields.json', **params)
    obs_fields = response.json()
    obs_fields = convert_all_timestamps(obs_fields)
    return {'results': obs_fields}


def put_observation_field_values(
    observation_id: int,
    observation_field_id: int,
    value: Any,
    access_token: str,
    **kwargs,
) -> JsonResponse:
    # TODO: Also implement a put_or_update_observation_field_values() that deletes then recreates the field_value?
    # TODO: Return some meaningful exception if it fails because the field is already set.
    # TODO: It appears pushing the same value/pair twice in a row (but deleting it meanwhile via the UI)...
    # TODO: ...triggers an error 404 the second time (report to iNaturalist?)
    """Set an observation field (value) on an observation.
    Will fail if this observation field is already set for this observation.

    To find an `observation_field_id`, either user :py:func:`.get_observation_fields` or search
    on iNaturalist: https://www.inaturalist.org/observation_fields

    **API reference:** https://www.inaturalist.org/pages/api+reference#put-observation_field_values-id

    Example:
        >>> # First find an observation field by name, if the ID is unknown
        >>> response = get_observation_fields('vespawatch_id')
        >>> observation_field_id = response[0]['id']
        >>>
        >>> put_observation_field_values(
        >>>     observation_id=7345179,
        >>>     observation_field_id=observation_field_id,
        >>>     value=250,
        >>>     access_token=token,
        >>> )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/put_observation_field_value_result.json
                :language: javascript

    Args:
        observation_id: ID of the observation receiving this observation field value
        observation_field_id: ID of the observation field for this observation field value
        value: Value for the observation field
        access_token: The access token, as returned by :func:`get_access_token()`

    Returns:
        The newly updated field value record
    """
    json_body = {
        'observation_field_value': {
            'observation_id': observation_id,
            'observation_field_id': observation_field_id,
            'value': value,
        }
    }

    response = put(
        f'{API_V0_BASE_URL}/observation_field_values/{observation_field_id}',
        access_token=access_token,
        json=json_body,
        **kwargs,
    )
    return response.json()



"""Summary of documentation components:

* ``forge`` is used to define and reuse documentation for API request parameters
* ``apidoc`` is used to generate rst sources for **modules**
* ``autosummary`` + ``automodapi`` to generate rst sources for **packages, classes, and summaries**
* ``intersphinx`` is used to insert links to other projects' docs
* Jinja templates provide some additional customization:
    * Show a class summary for ``pyinaturalist.models`` package
    * Show a function summary for ``pyinaturalist.v*`` packages
    * Show a function summary for all modules except ``pyinaturalist.models.*``
* Some Sphinx builder hooks copy some content so it can be accessed relative to ``docs`` dir
* Some CSS adds collapsible drop-down container
* Additional CSS is added to customize function-level docs
* A version dropdown (to switch to docs for older versions) is added with ``sphinx-material``, and
  versions are listed in ``static/versions.json``

TODO:

* Customization for package-level docs that can be done with module docstrings instead of templates?
* Automatically generate list of versions for version dropdown?
"""
import sys
from os import makedirs, symlink
from os.path import abspath, dirname, exists, join
from shutil import copytree, rmtree


# Avoid a potential circular import in nbsphinx
try:
    import prompt_toolkit  # noqa
except ImportError:
    pass

DOCS_DIR = abspath(dirname(__file__))
MODULE_DOCS_DIR = join(DOCS_DIR, 'modules')
PROJECT_DIR = dirname(DOCS_DIR)
PACKAGE_DIR = join(PROJECT_DIR, 'pyinaturalist')

# Source paths and symlink paths for static content to include
IMAGE_DIR_SRC = join(DOCS_DIR, 'images')
IMAGE_DIR_SYMLINK = join(DOCS_DIR, 'docs', 'images')
DATA_DIR_SRC = join(PROJECT_DIR, 'test', 'sample_data')
DATA_DIR_SYMLINK = join(DOCS_DIR, 'sample_data')
NOTEBOOK_DIR_SRC = join(PROJECT_DIR, 'examples')
NOTEBOOK_DIR_COPY = join(DOCS_DIR, 'examples')

# Add project path so we can import our package
sys.path.insert(0, PROJECT_DIR)
from pyinaturalist import __version__  # noqa: E402

# General information about the project.
project = 'pyinaturalist'
copyright = '2021, Nicolas Noé, Jordan Cook'
needs_sphinx = '4.0'
master_doc = 'index'
source_suffix = ['.rst', '.md']
version = release = __version__
html_static_path = ['_static']
templates_path = ['_templates']

# Exclude the generated pyinaturalist.rst, which will just contain top-level __init__ info
# and add an extra level to the toctree
exclude_patterns = ['_build', 'modules/pyinaturalist.rst']

# Sphinx extension modules
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'sphinx_copybutton',
    'sphinx_material',
    'sphinxcontrib.apidoc',
    'm2r2',
    'nbsphinx',
]

nbsphinx_allow_errors = True

# Enable automatic links to other projects' Sphinx docs
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'requests': ('https://requests.readthedocs.io/en/master/', None),
}

# Enable Google-style docstrings
napoleon_google_docstring = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False

# Strip prompt text when copying code blocks with copy button
copybutton_prompt_text = r'>>> |\.\.\. |\$ '
copybutton_prompt_is_regexp = True

# Move type hint info to function description instead of signature
autodoc_typehints = 'description'
set_type_checking_flag = True

# apidoc settings
apidoc_module_dir = PACKAGE_DIR
apidoc_output_dir = 'modules'
apidoc_excluded_paths = ['api_docs/*', 'node_api.py', 'rest_api.py']
apidoc_extra_args = ['--templatedir=_templates']
apidoc_module_first = True
apidoc_separate_modules = True
apidoc_toc_file = False

# autosummary + automodapi settings
automodapi_inheritance_diagram = False
automodsumm_inherited_members = False
autosummary_generate = True
autosummary_generate_overwrite = False
autosummary_imported_members = False
numpydoc_show_class_members = False

# HTML theme settings
html_theme = 'sphinx_material'
html_show_sphinx = False
html_theme_options = {
    'color_primary': 'blue-grey',
    'color_accent': 'teal',
    'globaltoc_depth': 3,
    'globaltoc_includehidden': False,
    'master_doc': False,
    'nav_title': project,
    'repo_url': 'https://github.com/niconoe/pyinaturalist',
    'repo_name': project,
    'version_dropdown': True,
    'version_json': '_static/versions.json',
}
html_sidebars = {'**': ['logo-text.html', 'globaltoc.html', 'localtoc.html', 'searchbox.html']}
pygments_style = 'friendly'

# TODO If/when dark mode support is added for sphinx-material
# pygments_style = 'material'

# Favicon & sidebar logo
# html_logo = 'logo.jpg'
html_favicon = join('images', 'favicon.ico')


def setup(app):
    """Run some additional steps after the Sphinx builder is initialized.

    In addition to basic things like adding static resources, this allows us to run any custom
    behavior that would otherwise go in the Makefile, so the readthedocs builder will behave the
    same as building the docs manually.

    Reference:
        * https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx-core-events
        * https://docs.readthedocs.io/en/stable/builds.html
        * https://github.com/sphinx-contrib/apidoc
    """
    app.connect('builder-inited', setup_external_files)
    app.connect('builder-inited', patch_automodapi)
    app.add_css_file('style.css')
    app.add_css_file('collapsible_container.css')


def setup_external_files(app):
    """Create symlinks and copies of files outside docs directory so they can be accessed by Sphinx
    directives.
    """
    make_symlink(IMAGE_DIR_SRC, IMAGE_DIR_SYMLINK)
    make_symlink(DATA_DIR_SRC, DATA_DIR_SYMLINK)
    # Unfortunately this can't be symlinked; nbsphinx will insert image links relative to this dir
    rmtree(NOTEBOOK_DIR_COPY, ignore_errors=True)
    copytree(NOTEBOOK_DIR_SRC, NOTEBOOK_DIR_COPY)


def make_symlink(src, dest):
    makedirs(dirname(dest), exist_ok=True)
    if not exists(dest):
        symlink(src, dest)


def patch_automodapi(app):
    """Monkey-patch the automodapi extension to exclude imported members.

    https://github.com/astropy/sphinx-automodapi/blob/master/sphinx_automodapi/automodsumm.py#L135
    """
    from sphinx_automodapi import automodsumm
    from sphinx_automodapi.utils import find_mod_objs

    automodsumm.find_mod_objs = lambda *args: find_mod_objs(args[0], onlylocals=True)



from pyinaturalist.api_docs import document_request_params
from pyinaturalist.api_docs import templates as docs
from pyinaturalist.constants import (
    NODE_OBS_ORDER_BY_PROPERTIES,
    HistogramResponse,
    IntOrStr,
    JsonResponse,
)
from pyinaturalist.converters import (
    convert_all_coordinates,
    convert_all_timestamps,
    convert_histogram,
    convert_observation_timestamps,
)
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.pagination import add_paginate_all
from pyinaturalist.request_params import validate_multiple_choice_param
from pyinaturalist.v1 import get_v1


def get_observation(observation_id: int, **params) -> JsonResponse:
    """Get details about a single observation by ID

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_id

    Example:

        >>> response = get_observation(16227955)
        >>> pprint(response)
        [16227955] [493595] Species: Lixus bardanae observed on 2018-09-05 14:06:00+01:00 by niconoe at 54 rue des Badauds

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation.py

    Args:
        observation_id: Get the observation with this ID. Only a single value is allowed.

    Returns:
        A dict with details on the observation

    Raises:
        :py:exc:`.ObservationNotFound` If an invalid observation is specified
    """

    response = get_observations(id=observation_id, **params)
    if response['results']:
        return convert_observation_timestamps(response['results'][0])
    raise ObservationNotFound()


@document_request_params([*docs._get_observations, docs._observation_histogram])
def get_observation_histogram(**params) -> HistogramResponse:
    """Search observations and return histogram data for the given time interval

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_histogram

    **Notes:**

    * Search parameters are the same as :py:func:`.get_observations()`, with the addition of
      ``date_field`` and ``interval``.
    * ``date_field`` may be either 'observed' (default) or 'created'.
    * Observed date ranges can be filtered by parameters ``d1`` and ``d2``
    * Created date ranges can be filtered by parameters ``created_d1`` and ``created_d2``
    * ``interval`` may be one of: 'year', 'month', 'week', 'day', 'hour', 'month_of_year', or
      'week_of_year'; spaces are also allowed instead of underscores, e.g. 'month of year'.
    * The year, month, week, day, and hour interval options will set default values for ``d1`` and
      ``created_d1``, to limit the number of groups returned. You can override those values if you
      want data from a longer or shorter time span.
    * The 'hour' interval only works with ``date_field='created'``

    Example:

        Get observations per month during 2020 in Austria (place ID 8057)

        >>> response = get_observation_histogram(
        >>>     interval='month',
        >>>     d1='2020-01-01',
        >>>     d2='2020-12-31',
        >>>     place_id=8057,
        >>> )

        .. admonition:: Example Response (observations per month of year)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_histogram_month_of_year.py

        .. admonition:: Example Response (observations per month)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_histogram_month.py

        .. admonition:: Example Response (observations per day)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_histogram_day.py

    Returns:
        Dict of ``{time_key: observation_count}``. Keys are ints for 'month of year' and\
        'week of year' intervals, and :py:class:`~datetime.datetime` objects for all other intervals.
    """
    response = get_v1('observations/histogram', **params)
    return convert_histogram(response.json())


@document_request_params([*docs._get_observations, docs._pagination, docs._only_id])
@add_paginate_all(method='id')
def get_observations(**params) -> JsonResponse:
    """Search observations.

    **API reference:** http://api.inaturalist.org/v1/docs/#!/Observations/get_observations

    Example:

        Get observations of Monarch butterflies with photos + public location info,
        on a specific date in the provice of Saskatchewan, CA (place ID 7953):

        >>> response = get_observations(
        >>>     taxon_name='Danaus plexippus',
        >>>     created_on='2020-08-27',
        >>>     photos=True,
        >>>     geo=True,
        >>>     geoprivacy='open',
        >>>     place_id=7953,
        >>> )

        Get basic info for observations in response:

        >>> pprint(response)
        '[57754375] Species: Danaus plexippus (Monarch) observed by samroom on 2020-08-27 at Railway Ave, Wilcox, SK'
        '[57707611] Species: Danaus plexippus (Monarch) observed by ingridt3 on 2020-08-26 at Michener Dr, Regina, SK'

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations_node.py

    Returns:
        Response dict containing observation records
    """
    validate_multiple_choice_param(params, 'order_by', NODE_OBS_ORDER_BY_PROPERTIES)
    response = get_v1('observations', **params)

    observations = response.json()
    observations['results'] = convert_all_coordinates(observations['results'])
    observations['results'] = convert_all_timestamps(observations['results'])

    return observations


@document_request_params([*docs._get_observations, docs._pagination])
@add_paginate_all(method='page')
def get_observation_species_counts(**params) -> JsonResponse:
    """Get all species (or other 'leaf taxa') associated with observations matching the search
    criteria, and the count of observations they are associated with.
    **Leaf taxa** are the leaves of the taxonomic tree, e.g., species, subspecies, variety, etc.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_species_counts

    Example:
        >>> response = get_observation_species_counts(user_login='my_username', quality_grade='research')
        >>> pprint(response)
        [62060] Species: Palomena prasina (Green Shield Bug): 10
        [84804] Species: Graphosoma italicum (European Striped Shield Bug): 8
        [55727] Species: Cymbalaria muralis (Ivy-leaved toadflax): 3
        ...

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_species_counts.py

    Returns:
        Response dict containing taxon records with counts
    """
    response = get_v1('observations/species_counts', **params)
    return response.json()


@document_request_params([*docs._get_observations, docs._pagination])
def get_observation_observers(**params) -> JsonResponse:
    """Get observers of observations matching the search criteria and the count of
    observations and distinct taxa of rank species they have observed.

    Notes:
        * Options for ``order_by`` are 'observation_count' (default) or 'species_count'
        * This endpoint will only return up to 500 results
        * See this issue for more details: https://github.com/inaturalist/iNaturalistAPI/issues/235

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_observers

    Example:
        >>> response = get_observation_observers(place_id=72645, order_by='species_count')
        >>> pprint(response, align=True)
        [1566366 ] fossa1211
        [674557  ] schurchin
        [5813    ] fluffberger (Fluff Berger)


        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_observers_ex_results.json
                :language: JSON

    Returns:
        Response dict of observers
    """
    params.setdefault('per_page', 500)
    response = get_v1('observations/observers', **params)
    return response.json()


@document_request_params([*docs._get_observations, docs._pagination])
def get_observation_identifiers(**params) -> JsonResponse:
    """Get identifiers of observations matching the search criteria and the count of
    observations they have identified. By default, results are sorted by ID count in descending.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_identifiers

    Note: This endpoint will only return up to 500 results.

    Example:
        >>> response = get_observation_identifiers(place_id=72645)
        >>> pprint(response)
        [409010  ] jdoe42 (Jane Doe)
        [691216  ] jbrown252 (James Brown)
        [3959037 ] tnsparkleberry

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_identifiers_ex_results.json
                :language: JSON

    Returns:
        Response dict of identifiers
    """
    params.setdefault('per_page', 500)
    response = get_v1('observations/identifiers', **params)
    return response.json()


@add_paginate_all(method='page')
def get_observation_taxonomy(user_id: IntOrStr, **params) -> JsonResponse:
    """Get observation counts for all taxa in a full taxonomic tree. In the web UI, these are used
    for life lists.

    Args:
        user_id: iNaturalist user ID or username

    Example:
        >>> response = get_observation_taxonomy(user_id='my_username')
        ...

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_taxonomy.json
                :language: JSON

    Returns:
        Response dict containing taxon records with counts
    """
    response = get_v1('observations/taxonomy', user_id=user_id, **params)
    return response.json()



#!/usr/bin/env python
# flake8: noqa: F401, F403
"""Sample response data of every type to experiment with"""
import sys

from pyinaturalist import PROJECT_DIR, format_table, pprint
from pyinaturalist.models import *

# Import sample response JSON used for unit tests
sys.path.insert(0, PROJECT_DIR)
from test.sample_data import *

# Sample model objects
annotation = Annotation.from_json(j_annotation_1)
comment = Comment.from_json(j_comment_1)
controlled_term = ControlledTerm.from_json(j_controlled_term_1)
identification = Identification.from_json(j_identification_1)
life_list = LifeList.from_json(j_life_list)
observation = Observation.from_json(j_observation_1)
observation_with_ofvs = Observation.from_json(j_observation_3_ofvs)
observation_field = ObservationField.from_json(j_obs_field_1)
ofv = ObservationFieldValue.from_json(j_ofv_1_numeric)
photo = Photo.from_json(j_photo_1)
photo_partial = Photo.from_json(j_photo_2_partial)
place = Place.from_json(j_place_1)
places_nearby = Place.from_json_list(j_places_nearby)
project = Project.from_json(j_project_1)
search_results = SearchResult.from_json_list(j_search_results)
taxon = Taxon.from_json(j_taxon_1)
taxon_partial = Taxon.from_json(j_taxon_2_partial)
taxon_counts = TaxonCounts.from_json(j_obs_species_counts)
user = User.from_json(j_user_1)

# Sample tables
tables = [
    format_table(observation_with_ofvs.annotations),
    format_table(observation.comments),
    format_table(ControlledTerm.from_json_list(j_controlled_terms)),
    format_table(observation.identifications),
    format_table(life_list),
    format_table([observation, observation_with_ofvs]),
    format_table(ObservationField.from_json_list(j_obs_fields)),
    format_table(observation_with_ofvs.ofvs),
    format_table([photo, photo_partial, observation.photos[0]]),
    format_table(places_nearby),
    format_table([project, project, project]),
    format_table(search_results),
    format_table(taxon_counts),
    format_table([taxon, observation_with_ofvs.taxon, observation.taxon]),
    format_table(User.from_json_list(j_users)),
]


if __name__ == '__main__':
    for table in tables:
        pprint(table)



#!/usr/bin/env python3
"""
An example of getting photo metadata from observation photos, which is currently not available in
the API and must be done by web scraping.

Note that web scraping in general is not very reliable and is prone to breakage, so this script
may or may not work without modification. See ``examples/sample_data/photo_info.html`` for an
example of photo info HTML at time of writing.

Also note that photo metadata is only visible to logged in users, so an access token is required.
For more details on authentication, see:

* https://www.inaturalist.org/pages/api+reference#auth
* https://pyinaturalist.readthedocs.io/en/latest/general_usage.html#authentication

Extra dependencies:
    ``pip install beautifulsoup4``
"""
from pprint import pprint

import requests
from bs4 import BeautifulSoup

from pyinaturalist import get_access_token, get_observation

IGNORE_ATTRIBUTES = ['Associated observations', 'Sizes']
PHOTO_INFO_BASE_URL = 'https://www.inaturalist.org/photos'


def get_photo_metadata(photo_url, access_token):
    """Scrape content from a photo info URL, and attempt to get its metadata"""
    print(f'Fetching {photo_url}')
    photo_page = requests.get(photo_url, headers={'Authorization': f'Bearer {access_token}'})
    soup = BeautifulSoup(photo_page.content, 'html.parser')
    table = soup.find(id='wrapper').find_all('table')[1]

    metadata = {}
    for row in table.find_all('tr'):
        key = row.find('th').text.strip()
        value = row.find('td').text.strip()
        if value and key not in IGNORE_ATTRIBUTES:
            metadata[key] = value
    return metadata


def get_observation_photo_metadata(observation_id, access_token):
    """Attempt to scrape metadata from all photo info pages associated with an observation"""
    print(f'Fetching observation {observation_id}')
    obs = get_observation(observation_id)
    photo_ids = [photo['id'] for photo in obs.get('photos', [])]
    photo_urls = [f'{PHOTO_INFO_BASE_URL}/{id}' for id in photo_ids]
    print(f'{len(photo_urls)} photo URL(s) found')
    return [get_photo_metadata(url, access_token) for url in photo_urls]


# !! Replace values here !!
if __name__ == '__main__':
    observation_id = 99
    access_token = get_access_token(
        username='',
        password='',
        app_id='',
        app_secret='',
    )
    all_metadata = get_observation_photo_metadata(observation_id, access_token)
    pprint(all_metadata, indent=4)



from pyinaturalist.constants import JsonResponse
from pyinaturalist.exceptions import TaxonNotFound
from pyinaturalist.v1 import get_v1


def get_controlled_terms(taxon_id: int = None, **params) -> JsonResponse:
    """List controlled terms and their possible values.
    A taxon ID can optionally be provided to show only terms that are valid for that taxon.
    Otherwise, all controlled terms will be returned.

    **API reference:**

    * https://api.inaturalist.org/v1/docs/#!/Controlled_Terms/get_controlled_terms
    * https://api.inaturalist.org/v1/docs/#!/Controlled_Terms/get_controlled_terms_for_taxon

    Example:

        >>> response = get_controlled_terms()
        >>> pprint(response)
        1: Life Stage
            2: Adult
            3: Teneral
            4: Pupa
        ...

        .. admonition:: Example Response (all terms)
            :class: toggle

            .. literalinclude:: ../sample_data/get_controlled_terms.json
                :language: JSON

        .. admonition:: Example Response (for a specific taxon)
            :class: toggle

            .. literalinclude:: ../sample_data/get_controlled_terms_for_taxon.json
                :language: JSON
    Args:
        taxon_id: ID of taxon to get controlled terms for

    Returns:
        A dict containing details on controlled terms and their values

    Raises:
        :py:exc:`.TaxonNotFound` If an invalid taxon_id is specified
    """
    # This is actually two endpoints, but they are so similar it seems best to combine them
    endpoint = 'controlled_terms/for_taxon' if taxon_id else 'controlled_terms'
    response = get_v1(endpoint, params={'taxon_id': taxon_id}, **params)

    # controlled_terms/for_taxon returns a 422 if the specified taxon does not exist
    if response.status_code in (404, 422):
        raise TaxonNotFound
    return response.json()



from datetime import datetime

from pyinaturalist.constants import API_V1_BASE_URL
from pyinaturalist.v1 import search
from test.conftest import load_sample_data


def test_search(requests_mock):
    """Simulate /search results with one of each record type"""
    requests_mock.get(
        f'{API_V1_BASE_URL}/search',
        json=load_sample_data('get_search.json'),
        status_code=200,
    )
    response = search([8348, 6432])
    taxon_result = response['results'][0]
    place_result = response['results'][1]
    project_result = response['results'][2]
    user_result = response['results'][3]

    assert all([isinstance(result['score'], float) for result in response['results']])

    # Mainly just need to test type conversions
    assert taxon_result['type'] == 'Taxon'
    assert isinstance(taxon_result['record']['created_at'], datetime)

    assert place_result['type'] == 'Place'
    assert isinstance(place_result['record']['location'][0], float)
    assert isinstance(place_result['record']['location'][1], float)

    assert project_result['type'] == 'Project'
    assert isinstance(project_result['record']['last_post_at'], datetime)

    assert user_result['type'] == 'User'
    assert isinstance(user_result['record']['created_at'], datetime)



{
    'total_results': 3,
    'page': 1,
    'per_page': 10,
    'results': [
        {
            'observations_count': 235680,
            'taxon_schemes_count': 2,
            'ancestry': '48460/1/47120/372739/47158/184884/47201/124417/326777/48740',
            'is_active': True,
            'flag_counts': {'unresolved': 0, 'resolved': 1},
            'wikipedia_url': 'http://en.wikipedia.org/wiki/Vespidae',
            'current_synonymous_taxon_ids': None,
            'iconic_taxon_id': 47158,
            'rank_level': 30,
            'taxon_changes_count': 0,
            'atlas_id': None,
            'complete_species_count': None,
            'parent_id': 48740,
            'name': 'Vespidae',
            'rank': 'family',
            'extinct': False,
            'id': 52747,
            'default_photo': {
                'square_url': 'https://static.inaturalist.org/photos/13765/square.jpg?1545358964',
                'attribution': '(c) CÃ©cile Bassaglia, some rights reserved (CC BY-NC-SA)',
                'flags': [],
                'medium_url': 'https://static.inaturalist.org/photos/13765/medium.jpg?1545358964',
                'id': 13765,
                'license_code': 'cc-by-nc-sa',
                'original_dimensions': {'width': 1024, 'height': 679},
                'url': 'https://static.inaturalist.org/photos/13765/square.jpg?1545358964',
            },
            'ancestor_ids': [
                48460,
                1,
                47120,
                372739,
                47158,
                184884,
                47201,
                124417,
                326777,
                48740,
                52747,
            ],
            'matched_term': 'Vespidae',
            'iconic_taxon_name': 'Insecta',
            'preferred_common_name': 'Hornets, Paper Wasps, Potter Wasps, and Allies',
        },
        {
            'observations_count': 74782,
            'taxon_schemes_count': 1,
            'ancestry': '48460/1/47120/372739/47158/184884/47201/124417/326777/48740/52747',
            'is_active': True,
            'flag_counts': {'unresolved': 1, 'resolved': 0},
            'wikipedia_url': 'http://en.wikipedia.org/wiki/Vespinae',
            'current_synonymous_taxon_ids': None,
            'iconic_taxon_id': 47158,
            'rank_level': 27,
            'taxon_changes_count': 0,
            'atlas_id': None,
            'complete_species_count': None,
            'parent_id': 52747,
            'name': 'Vespinae',
            'rank': 'subfamily',
            'extinct': False,
            'id': 84738,
            'default_photo': {
                'square_url': 'https://static.inaturalist.org/photos/11374710/square.jpg?1508777024',
                'attribution': '(c) Arnold Wijker, some rights reserved (CC BY-NC)',
                'flags': [],
                'medium_url': 'https://static.inaturalist.org/photos/11374710/medium.jpg?1508777024',
                'id': 11374710,
                'license_code': 'cc-by-nc',
                'original_dimensions': {'width': 800, 'height': 639},
                'url': 'https://static.inaturalist.org/photos/11374710/square.jpg?1508777024',
            },
            'ancestor_ids': [
                48460,
                1,
                47120,
                372739,
                47158,
                184884,
                47201,
                124417,
                326777,
                48740,
                52747,
                84738,
            ],
            'matched_term': 'Vespinae',
            'iconic_taxon_name': 'Insecta',
            'preferred_common_name': 'Hornets and Yellowjackets',
        },
        {
            'observations_count': 610,
            'taxon_schemes_count': 1,
            'ancestry': '48460/1/47120/372739/47158/184884/47208/71130/372870/48311/53849/204079/53850',
            'is_active': True,
            'flag_counts': {'unresolved': 0, 'resolved': 0},
            'wikipedia_url': 'http://en.wikipedia.org/wiki/Nicrophorus_vespillo',
            'current_synonymous_taxon_ids': None,
            'iconic_taxon_id': 47158,
            'rank_level': 10,
            'taxon_changes_count': 0,
            'atlas_id': None,
            'complete_species_count': None,
            'parent_id': 53850,
            'name': 'Nicrophorus vespillo',
            'rank': 'species',
            'extinct': False,
            'id': 131878,
            'default_photo': {
                'square_url': 'https://static.inaturalist.org/photos/24573164/square.jpg?1536478624',
                'attribution': '(c) Nikolai Vladimirov, some rights reserved (CC BY-NC)',
                'flags': [],
                'medium_url': 'https://static.inaturalist.org/photos/24573164/medium.jpg?1536478624',
                'id': 24573164,
                'license_code': 'cc-by-nc',
                'original_dimensions': {'width': 2048, 'height': 1365},
                'url': 'https://static.inaturalist.org/photos/24573164/square.jpg?1536478624',
            },
            'ancestor_ids': [
                48460,
                1,
                47120,
                372739,
                47158,
                184884,
                47208,
                71130,
                372870,
                48311,
                53849,
                204079,
                53850,
                131878,
            ],
            'matched_term': 'Nicrophorus vespillo',
            'iconic_taxon_name': 'Insecta',
            'preferred_common_name': 'Vespillo Burying Beetle',
        },
    ],
}



from datetime import datetime
from typing import Dict, List, Optional

from attr import field, fields_dict

from pyinaturalist.constants import (
    ICONIC_EMOJI,
    ICONIC_TAXA_BASE_URL,
    INAT_BASE_URL,
    RANKS,
    JsonResponse,
    TableRow,
)
from pyinaturalist.models import (
    BaseModel,
    BaseModelCollection,
    LazyProperty,
    Photo,
    Place,
    User,
    datetime_attr,
    define_model,
    kwarg,
)
from pyinaturalist.v1 import get_taxa_by_id


@define_model
class ConservationStatus(BaseModel):
    authority: str = kwarg
    created_at: Optional[datetime] = datetime_attr
    description: str = kwarg
    geoprivacy: str = kwarg
    iucn: int = kwarg
    place_id: int = kwarg
    source_id: int = kwarg
    status: str = kwarg  # Enum
    status_name: str = kwarg
    taxon_id: int = kwarg
    updated_at: Optional[datetime] = datetime_attr
    updater_id: int = kwarg
    url: str = kwarg

    # Lazy-loaded nested model objects
    place: property = LazyProperty(Place.from_json)
    updater: property = LazyProperty(User.from_json)
    user: property = LazyProperty(User.from_json)


@define_model
class EstablishmentMeans(BaseModel):
    """The establishment means for a taxon in a given location"""

    establishment_means: str = kwarg  # Enum: introduced, etc.
    id: int = kwarg
    place: property = LazyProperty(Place.from_json_list)

    def __str__(self) -> str:
        return self.establishment_means


@define_model
class Taxon(BaseModel):
    """An iNaturalist taxon, based on the schema of
    `GET /taxa <https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa>`_.

    Can be constructed from either a full or partial JSON record. Examples of partial records
    include nested ``ancestors``, ``children``, and results from :py:func:`.get_taxa_autocomplete`.
    """

    complete_rank: str = kwarg  #: Complete or "leaf taxon" rank, e.g. species or subspecies
    complete_species_count: int = kwarg  #: Total number of species descended from this taxon
    created_at: Optional[datetime] = datetime_attr  #: When the taxon was added to iNaturalist
    extinct: bool = kwarg  #: Indicates if the taxon is extinct
    iconic_taxon_id: int = field(default=0)  #: ID of the iconic taxon or taxon "category"
    iconic_taxon_name: str = field(default='unknown')  #: Name of the iconic taxon or taxon "category"
    id: int = kwarg  #: Taxon ID
    is_active: bool = kwarg  #: Indicates if the taxon is active (and not renamed, moved, etc.)
    listed_taxa_count: int = kwarg  #: Number of listed taxa from this taxon + descendants
    name: str = kwarg  #: Taxon name; contains full scientific name at species level and below
    observations_count: int = kwarg  #: Total number of observations of this taxon and its descendants
    parent_id: int = kwarg  #: Taxon ID of immediate ancestor
    preferred_common_name: str = field(default='')  #: Common name for the preferred place, if any
    preferred_establishment_means: str = kwarg  #: Establishment means for this taxon + preferred place
    rank_level: int = kwarg  #: Rank number for easier comparison between ranks (kingdom=higest)
    rank: str = kwarg  #: Taxon rank (species, genus, etc.)
    taxon_changes_count: int = kwarg  #: Number of curator changes to this taxon
    taxon_schemes_count: int = kwarg  #: Taxon schemes that include this taxon
    wikipedia_summary: str = kwarg  #: Taxon summary from Wikipedia article, if available
    wikipedia_url: str = kwarg  #: URL to Wikipedia article for the taxon, if available

    # Nested collections
    ancestor_ids: List[int] = field(factory=list)  #: Taxon IDs of ancestors, from highest rank to lowest
    current_synonymous_taxon_ids: List[int] = field(factory=list)  #: Taxa that are accepted synonyms
    listed_taxa: List = field(factory=list)  #: Listed taxon IDs

    # Lazy-loaded nested model objects
    ancestors: property = LazyProperty(BaseModel.from_json_list)
    children: property = LazyProperty(BaseModel.from_json_list)
    conservation_status: property = LazyProperty(ConservationStatus.from_json)
    conservation_statuses: property = LazyProperty(ConservationStatus.from_json_list)
    default_photo: property = LazyProperty(Photo.from_json)
    establishment_means: property = LazyProperty(EstablishmentMeans.from_json)
    taxon_photos: property = LazyProperty(Photo.from_json_list)

    # Unused attributes
    # atlas_id: int = kwarg
    # flag_counts: Dict[str, int] = field(factory=dict)  # {"unresolved": 1, "resolved": 2}
    # min_species_ancestry: str = kwarg  #: Used internally by iNaturalist for Elasticsearch aggregations
    # min_species_taxon_id: int = kwarg
    # photos_locked: bool = kwarg
    # universal_search_rank: int = kwarg

    @classmethod
    def from_sorted_json_list(cls, value: JsonResponse) -> List['Taxon']:
        """Sort Taxon objects by rank then by name"""
        taxa = cls.from_json_list(value)
        taxa.sort(key=_get_rank_name_idx)
        return taxa  # type: ignore

    @property
    def ancestry(self) -> str:
        """String containing either ancestor names (if available) or IDs"""
        tokens = [t.name for t in self.ancestors] if self.ancestors else self.ancestor_ids
        return ' | '.join([str(t) for t in tokens])

    @property
    def emoji(self) -> str:
        """Get an emoji representing the iconic taxon"""
        return ICONIC_EMOJI.get(self.iconic_taxon_id, '❓')

    @property
    def full_name(self) -> str:
        """Taxon rank, scientific name, and common name (if available).

        Example:

            >>> taxon.full_name
            'Genus: Physcia (Rosette Lichens)'

        """
        if not self.name and not self.rank:
            return 'unknown taxon'
        if not self.name:
            name = str(self.id)
        else:
            common_name = self.preferred_common_name
            name = self.name + (f' ({common_name})' if common_name else '')

        return f'{self.rank.title()}: {name}'

    @property
    def icon_url(self) -> str:
        """Get a URL to the iconic taxon's icon"""
        return f'{ICONIC_TAXA_BASE_URL}/{self.iconic_taxon_name.lower()}-75px.png'

    @property
    def parent(self) -> 'Taxon':
        """Get immediate parent, if any"""
        return self.ancestors[-1] if self.ancestors else None

    @property
    def url(self) -> str:
        """Info URL on iNaturalist.org"""
        return f'{INAT_BASE_URL}/taxa/{self.id}'

    @classmethod
    def from_id(cls, id: int) -> 'Taxon':
        """Lookup and create a new Taxon object by ID"""
        r = get_taxa_by_id(id)
        return cls.from_json(r['results'][0])  # type: ignore

    def load_full_record(self):
        """Update this Taxon with full taxon info, including ancestors + children"""
        t = Taxon.from_id(self.id)
        for key in fields_dict(Taxon).keys():
            key = key.lstrip('_')  # Use getters/setters for LazyProperty instead of temp attrs
            setattr(self, key, getattr(t, key))

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Rank': self.rank,
            'Scientific name': f'{self.emoji} {self.name}',
            'Common name': self.preferred_common_name,
        }

    def __str__(self) -> str:
        return f'[{self.id}] {self.full_name}' if self.name else self.full_name


@define_model
class TaxonCount(Taxon):
    """A taxon with an associated count, used in a :py:class:`.TaxonCounts` collection"""

    count: int = field(default=0)  #: Number of observations of this taxon

    @classmethod
    def from_json(cls, value: JsonResponse, user_id: int = None, **kwargs) -> 'TaxonCount':
        """Flatten out count + taxon fields into a single-level dict before initializing"""
        if 'results' in value:
            value = value['results']
        if 'taxon' in value:
            value.update(value.pop('taxon'))
        return super(TaxonCount, cls).from_json(value)  # type: ignore

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Rank': self.rank,
            'Name': self.name,
            'Count': self.count,
        }

    def __str__(self) -> str:
        return f'[{self.id}] {self.full_name}: {self.count}'


@define_model
class TaxonCounts(BaseModelCollection):
    """A collection of taxa with an associated counts. Used with
    `GET /observations/species_counts <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_species_counts>`_.
    as well as :py:class:`.LifeList`.
    """

    taxa: List[TaxonCount] = field(factory=list, converter=TaxonCount.from_json_list)
    _taxon_counts: Dict[int, int] = field(default=None, init=False, repr=False)

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'TaxonCounts':
        if 'results' in value:
            value = value['results']
        if 'taxa' not in value:
            value = {'taxa': value}
        return super(TaxonCounts, cls).from_json(value)  # type: ignore

    def count(self, taxon_id: int) -> int:
        """Get an observation count for the specified taxon and its descendants, and handle unlisted taxa.
        **Note:** ``-1`` can be used an alias for ``count_without_taxon``.
        """
        # Make and cache an index of taxon IDs and observation counts
        if self._taxon_counts is None:
            self._taxon_counts = {t.id: t.descendant_obs_count for t in self.taxa}
            self._taxon_counts[-1] = self.count_without_taxon
        return self._taxon_counts.get(taxon_id, 0)

    def __str__(self) -> str:
        return '\n'.join([str(taxon) for taxon in self.taxa])


# Since these use Taxon classmethods, they must be added after Taxon is defined
Taxon.ancestors = LazyProperty(Taxon.from_json_list, 'ancestors')
Taxon.children = LazyProperty(Taxon.from_sorted_json_list, 'children')


def _get_rank_name_idx(taxon):
    return _get_rank_idx(taxon.rank), taxon.name


def _get_rank_idx(rank: str) -> int:
    return RANKS.index(rank) if rank in RANKS else 0



# TODO: Method to preview image in Jupyter
from typing import Optional, Tuple

from attr import field

from pyinaturalist.constants import CC_LICENSES, PHOTO_INFO_BASE_URL, PHOTO_SIZES, TableRow
from pyinaturalist.converters import format_dimensions, format_license
from pyinaturalist.models import BaseModel, define_model, kwarg


@define_model
class Photo(BaseModel):
    """An observation photo, based on the schema of photos from
    `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    attribution: str = kwarg
    id: int = kwarg
    license_code: str = field(converter=format_license, default=None)  # Enum
    original_dimensions: Tuple[int, int] = field(converter=format_dimensions, default=(0, 0))
    url: str = kwarg
    _url_format: str = field(init=False, repr=False, default=None)

    # Unused attributes
    # flags: List = field(factory=list)

    def __attrs_post_init__(self):
        if not self.url:
            return

        # Get a URL format string to get different photo sizes. Note: default URL may be any size.
        for size in PHOTO_SIZES:
            if f'{size}.' in self.url:
                self._url_format = self.url.replace(size, '{size}')

    @property
    def dimensions_str(self) -> str:
        return f'{self.original_dimensions[0]}x{self.original_dimensions[1]}'

    @property
    def has_cc_license(self) -> bool:
        """Determine if this photo has a Creative Commons license"""
        return self.license_code in CC_LICENSES

    @property
    def info_url(self) -> str:
        return f'{PHOTO_INFO_BASE_URL}/{self.id}'

    @property
    def large_url(self) -> Optional[str]:
        return self.url_size('large')

    @property
    def medium_url(self) -> Optional[str]:
        return self.url_size('medium')

    @property
    def original_url(self) -> Optional[str]:
        return self.url_size('original')

    @property
    def small_url(self) -> Optional[str]:
        return self.url_size('small')

    @property
    def square_url(self) -> Optional[str]:
        return self.url_size('square')

    @property
    def thumbnail_url(self) -> Optional[str]:
        return self.url_size('square')

    def url_size(self, size: str) -> Optional[str]:
        size = size.replace('thumbnail', 'square').replace('thumb', 'square')
        if not self._url_format or size not in PHOTO_SIZES:
            return None
        return self._url_format.format(size=size)

    def open(self, size: str = 'large'):
        """Download and display the image with the system's default image viewer.
        Experimental / requires ``pillow``
        """
        import requests
        from PIL import Image

        url = self.url_size(size)
        if url:
            img = Image.open(requests.get(url, stream=True).raw)
            img.show()

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'License': self.license_code,
            'Dimensions': self.dimensions_str,
            'URL': self.original_url,
        }

    def __str__(self) -> str:
        return f'[{self.id}] {self.original_url} ({self.license_code}, {self.dimensions_str})'



from datetime import datetime
from dateutil.tz import tzutc

from pyinaturalist.constants import API_V1_BASE_URL
from pyinaturalist.v1 import get_projects, get_projects_by_id
from test.conftest import load_sample_data


def test_get_projects(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/projects',
        json=load_sample_data('get_projects.json'),
        status_code=200,
    )

    response = get_projects(q='invasive', lat=49.27, lng=-123.08, radius=400, order_by='distance')
    first_result = response['results'][0]

    assert response['total_results'] == len(response['results']) == 5
    assert first_result['id'] == 8291
    assert first_result['title'] == 'PNW Invasive Plant EDDR'
    assert first_result['is_umbrella'] is False
    assert len(first_result['user_ids']) == 33
    assert first_result['created_at'] == datetime(2016, 7, 20, 23, 0, 5, tzinfo=tzutc())
    assert first_result['updated_at'] == datetime(2020, 7, 28, 20, 9, 49, tzinfo=tzutc())


def test_get_projects_by_id(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/projects/8348,6432',
        json=load_sample_data('get_projects_by_id.json'),
        status_code=200,
    )
    response = get_projects_by_id([8348, 6432])
    first_result = response['results'][0]

    assert response['total_results'] == len(response['results']) == 2
    assert first_result['id'] == 8348
    assert first_result['title'] == 'Tucson High Native and Invasive Species Inventory'
    assert first_result['place_id'] == 96103
    assert first_result['location'] == [32.2264416406, -110.9617278383]
    assert first_result['created_at'] == datetime(2016, 7, 26, 23, 8, 47, tzinfo=tzutc())
    assert first_result['updated_at'] == datetime(2017, 9, 16, 1, 51, 1, tzinfo=tzutc())



import pytest
from datetime import datetime
from dateutil.tz import tzoffset, tzutc
from unittest.mock import patch

from pyinaturalist.constants import API_V1_BASE_URL
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.v1 import (
    get_observation,
    get_observation_histogram,
    get_observation_identifiers,
    get_observation_observers,
    get_observation_species_counts,
    get_observation_taxonomy,
    get_observations,
)
from test.conftest import load_sample_data


def test_get_observation(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_observation.json'),
        status_code=200,
    )

    observation = get_observation(16227955)
    assert observation['quality_grade'] == 'research'
    assert observation['id'] == 16227955
    assert observation['user']['login'] == 'niconoe'
    assert len(observation['photos']) == 2


def test_get_observation_histogram(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations/histogram',
        json=load_sample_data('get_observation_histogram_month.json'),
        status_code=200,
    )

    histogram = get_observation_histogram(
        interval='month', place_id=24, d1='2020-01-01', d2='2020-12-31'
    )
    assert len(histogram) == 12
    assert histogram[datetime(2020, 1, 1, 0, 0)] == 272
    assert all([isinstance(k, datetime) for k in histogram.keys()])


def test_get_observations(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_observations_node_page1.json'),
        status_code=200,
    )

    observations = get_observations(
        taxon_name='Danaus plexippus',
        created_on='2020-08-27',
        photos=True,
        geo=True,
        geoprivacy='open',
        place_id=7953,
    )
    first_result = observations['results'][0]

    assert observations['total_results'] == 2
    assert len(observations['results']) == 1
    assert first_result['taxon_geoprivacy'] == 'open'
    assert first_result['created_at'] == datetime(2020, 8, 27, 18, 0, 51, tzinfo=tzutc())
    assert first_result['observed_on'] == datetime(2020, 8, 27, 8, 57, 22, tzinfo=tzoffset('Etc/UTC', 0))
    assert first_result['taxon']['id'] == 48662
    assert len(first_result['place_ids']) == 13


@patch('pyinaturalist.pagination.sleep')
def test_get_observations__all_pages(sleep, requests_mock):
    page_1 = load_sample_data('get_observations_node_page1.json')
    page_2 = load_sample_data('get_observations_node_page2.json')

    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        [
            {'json': page_1, 'status_code': 200},
            {'json': page_2, 'status_code': 200},
        ],
    )

    observations = get_observations(id=[57754375, 57707611], page='all')

    assert len(observations['results']) == 2


def test_get_observation_observers(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations/observers',
        json=load_sample_data('get_observation_observers_node_page1.json'),
        status_code=200,
    )

    observers = get_observation_observers(place_id=125323)
    first_result = observers['results'][0]

    assert observers['total_results'] == 4
    assert len(observers['results']) == 2
    assert first_result['user']['spam'] is False
    assert first_result['user']['suspended'] is False


def test_get_observation_identifiers(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations/identifiers',
        json=load_sample_data('get_observation_identifiers_node_page1.json'),
        status_code=200,
    )

    identifiers = get_observation_identifiers(place_id=125323, iconic_taxa='Amphibia')
    first_result = identifiers['results'][0]

    assert identifiers['total_results'] == 6
    assert len(identifiers['results']) == 3
    assert first_result['user']['spam'] is False
    assert first_result['user']['suspended'] is False


def test_get_non_existent_observation(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_nonexistent_observation.json'),
        status_code=200,
    )
    with pytest.raises(ObservationNotFound):
        get_observation(99999999)


def test_get_observation_species_counts(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations/species_counts',
        json=load_sample_data('get_observation_species_counts.json'),
        status_code=200,
    )
    response = get_observation_species_counts(user_login='my_username', quality_grade='research')
    first_result = response['results'][0]

    assert first_result['count'] == 31
    assert first_result['taxon']['id'] == 48484
    assert first_result['taxon']['name'] == 'Harmonia axyridis'


@patch('pyinaturalist.pagination.sleep')
def test_get_observation_species_counts__all_pages(sleep, requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations/species_counts',
        [
            {
                'json': load_sample_data('get_all_observation_species_counts_page1.json'),
                'status_code': 200,
            },
            {
                'json': load_sample_data('get_all_observation_species_counts_page2.json'),
                'status_code': 200,
            },
        ],
    )
    response = get_observation_species_counts(
        user_login='my_username', quality_grade='research', page='all'
    )
    first_result = response['results'][0]
    last_result = response['results'][-1]

    assert len(response['results']) == 22
    assert first_result['count'] == 19
    assert first_result['taxon']['id'] == 27805
    assert first_result['taxon']['name'] == 'Notophthalmus viridescens'
    assert last_result['count'] == 1
    assert last_result['taxon']['id'] == 39556
    assert last_result['taxon']['name'] == 'Apalone spinifera'


def test_get_observation_species_counts__invalid_multiple_choice_params():
    with pytest.raises(ValueError):
        get_observation_species_counts(quality_grade='None', iconic_taxa='slime molds')


def test_get_observation_taxonomy(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations/taxonomy',
        json=load_sample_data('get_observation_taxonomy.json'),
        status_code=200,
    )
    response = get_observation_taxonomy(user_id=12345)
    first_result = response['results'][0]

    assert response['count_without_taxon'] == 4
    assert first_result['id'] == 1
    assert first_result['name'] == 'Animalia'
    assert first_result['descendant_obs_count'] == 3023



from datetime import datetime
from dateutil.tz import tzutc

from pyinaturalist.constants import API_V0_BASE_URL
from pyinaturalist.v0 import get_observation_fields, put_observation_field_values
from test.conftest import load_sample_data


def test_get_observation_fields(requests_mock):
    """get_observation_fields() work as expected (basic use)"""
    page_2_json_response = load_sample_data('get_observation_fields_page2.json')

    requests_mock.get(
        f'{API_V0_BASE_URL}/observation_fields.json?q=sex&page=2',
        json=page_2_json_response,
        status_code=200,
    )

    response = get_observation_fields(q='sex', page=2)
    first_result = response['results'][0]
    assert len(response['results']) == 3
    assert first_result['id'] == 5
    assert first_result['datatype'] == 'text'
    assert first_result['created_at'] == datetime(2012, 1, 23, 8, 12, 0, 138000, tzinfo=tzutc())
    assert first_result['updated_at'] == datetime(2018, 10, 16, 6, 47, 43, 975000, tzinfo=tzutc())


def test_get_observation_fields__all_pages(requests_mock):
    page_1_json_response = load_sample_data('get_observation_fields_page1.json')
    page_2_json_response = load_sample_data('get_observation_fields_page2.json')

    requests_mock.get(
        f'{API_V0_BASE_URL}/observation_fields.json?q=sex&page=1',
        json=page_1_json_response,
        status_code=200,
    )
    requests_mock.get(
        f'{API_V0_BASE_URL}/observation_fields.json?q=sex&page=2',
        json=page_2_json_response,
        status_code=200,
    )
    requests_mock.get(
        f'{API_V0_BASE_URL}/observation_fields.json?q=sex&page=3',
        json=[],
        status_code=200,
    )
    expected_results = page_1_json_response + page_2_json_response

    response = get_observation_fields(q='sex', page='all')
    first_result = response['results'][0]
    assert response['total_results'] == len(response['results']) == len(expected_results)
    assert first_result['created_at'] == datetime(2016, 5, 29, 16, 17, 8, 51000, tzinfo=tzutc())
    assert first_result['updated_at'] == datetime(2018, 1, 1, 1, 17, 56, 7000, tzinfo=tzutc())


def test_put_observation_field_values(requests_mock):
    requests_mock.put(
        f'{API_V0_BASE_URL}/observation_field_values/31',
        json=load_sample_data('put_observation_field_value_result.json'),
        status_code=200,
    )

    response = put_observation_field_values(
        observation_id=18166477,
        observation_field_id=31,  # Animal behavior
        value='fouraging',
        access_token='valid token',
    )

    assert response['id'] == 31
    assert response['observation_field_id'] == 31
    assert response['observation_id'] == 18166477
    assert response['value'] == 'fouraging'



#!/usr/bin/env python3
"""
An example of converting observation locations + metadata into GPX format.

Extra dependencies:
    ``pip install gpxpy``
"""
from logging import getLogger

from gpxpy.gpx import GPX, GPXTrack, GPXTrackPoint, GPXTrackSegment, GPXWaypoint

from pyinaturalist import format_observations, get_observations
from pyinaturalist.constants import JsonResponse, List
from pyinaturalist.response_format import convert_observation_timestamps

logger = getLogger(__name__)


def observations_to_gpx(
    observations: List[JsonResponse], output_file: str = "observations.gpx", track: bool = True
):
    """Convert a list of observations to a set of GPX waypoints or a GPX track

    Args:
        observations: JSON observations
        output_file: File path to write to
        track: Create an ordered GXP track; otherwise, create unordered GPX waypoints
    """
    gpx = GPX()
    logger.info(f"Converting {len(observations)} to GPX points")
    points = [observation_to_gpx_point(obs, track=track) for obs in observations]

    if track:
        gpx_track = GPXTrack()
        gpx.tracks.append(gpx_track)
        gpx_segment = GPXTrackSegment()
        gpx_track.segments.append(gpx_segment)
        gpx_segment.points = points
    else:
        gpx.waypoints = points

    # Save to file
    logger.info(f"Writing GPX data to {output_file}")
    with open(output_file, "w") as f:
        f.write(gpx.to_xml())


def observation_to_gpx_point(observation: JsonResponse, track: bool = True):
    """Convert a single observation to a GPX point

    Args:
        observation: JSON observation
        track: Indicates that this point is part of an ordered GXP track;
            otherwise, assume it is an unordered waypoint

    """
    logger.debug(f'Processing observation {observation["id"]}')
    # GeoJSON coordinates are ordered as `longitude, latitude`
    long, lat = observation["geojson"]["coordinates"]

    # Get medium-sized photo URL, if available; otherwise just use observation URL
    if observation["photos"]:
        link = observation["photos"][0]["url"].replace("square", "medium")
    else:
        link = observation["uri"]

    point_cls = GPXTrackPoint if track else GPXWaypoint
    point = point_cls(
        latitude=lat,
        longitude=long,
        time=convert_observation_timestamps(observation),
        comment=format_observations(observation),
    )
    point.description = observation["description"]
    point.link = link
    point.link_text = f'Observation {observation["id"]}'
    return point


if __name__ == "__main__":
    # Get first page of search results (for testing)
    search_params = {
        "project_id": 36883,  # ID of the 'Sugarloaf Ridge State Park' project
        "created_d1": "2020-01-01",  # Get observations from January 2020...
        "created_d2": "2020-09-30",  # ...through September 2020 (adjust these dates as needed)
        "geo": True,  # Only get observations with geospatial coordinates
        "geoprivacy": "open",  # Only get observations with public coordinates (not obscured/private)
        "page": "all",  # Paginate through all response pages
    }
    results = get_observations(**search_params)["results"]

    # Paginate through all search results (may take a long time for a large query)
    # results = get_observations(**search_params)

    # Convert and write to GPX file
    observations_to_gpx(results)
    # observations_to_tsp(results)



import pytest
from datetime import datetime, timedelta
from dateutil.tz import tzutc
from io import BytesIO
from unittest.mock import patch

from requests import HTTPError

from pyinaturalist.constants import API_V0_BASE_URL, OBSERVATION_FORMATS
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.v0 import (
    create_observation,
    delete_observation,
    get_observations,
    update_observation,
    upload_photos,
    upload_sounds,
)
from test.conftest import load_sample_data


def get_observations_response(converters):
    converters = converters.replace('widget', 'js')
    return load_sample_data(f'get_observations.{converters}')


@pytest.mark.parametrize('converters', OBSERVATION_FORMATS)
def test_get_observations(converters, requests_mock):
    """Test all supported observation data formats"""
    mock_response = get_observations_response(converters)
    key = 'json' if converters == 'json' else 'text'

    requests_mock.get(
        f'{API_V0_BASE_URL}/observations.{converters}',
        status_code=200,
        **{key: mock_response},
    )

    observations = get_observations(taxon_id=493595, converters=converters)

    # For JSON format, ensure type conversions were performed
    if converters == 'json':
        assert observations[0]['latitude'] == 50.646894
        assert observations[0]['longitude'] == 4.360086
        assert observations[0]['created_at'] == datetime(2018, 9, 5, 12, 31, 8, 48000, tzinfo=tzutc())
        assert observations[0]['updated_at'] == datetime(2018, 9, 22, 17, 19, 27, 80000, tzinfo=tzutc())
    else:
        assert observations == mock_response


@pytest.mark.parametrize('converters', ['geojson', 'yaml'])
def test_get_observations__invalid_format(converters):
    with pytest.raises(ValueError):
        get_observations(taxon_id=493595, converters=converters)


def test_create_observation(requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/observations.json',
        json=load_sample_data('create_observation_result.json'),
        status_code=200,
    )

    response = create_observation(
        species_guess='Pieris rapae', observed_on_string='2021-09-09', access_token='valid token'
    )
    assert len(response) == 1
    assert response[0]['latitude'] is None
    assert response[0]['taxon_id'] == 55626


@patch('pyinaturalist.v0.observations.post')
def test_create_observation__with_datetime(post, requests_mock):
    """Just test that request param aliases work. Datetimes are converted to strings separately in
    request_params.prepare_request().
    """
    requests_mock.post(
        f'{API_V0_BASE_URL}/observations.json',
        json=load_sample_data('create_observation_result.json'),
        status_code=200,
    )

    now = datetime.now()
    create_observation(species_guess='Pieris rapae', observed_on=now, access_token='valid token')
    request_params = post.call_args[1]['json']['observation']
    assert request_params['observed_on_string'] == now


@patch('pyinaturalist.v0.observations.upload_photos')
@patch('pyinaturalist.v0.observations.post')
def test_create_observation__with_photos(post, mock_upload_photos):
    create_observation(access_token='token', photos=['photo_1.jpg', 'photo_2.jpg'])

    request_params = post.call_args[1]['json']['observation']
    assert 'local_photos' not in request_params
    mock_upload_photos.assert_called_once()


@patch('pyinaturalist.v0.observations.upload_sounds')
@patch('pyinaturalist.v0.observations.post')
def test_create_observation__with_sounds(post, mock_upload_sounds):
    create_observation(access_token='token', sounds=['sound_1.mp3', 'sound_2.wav'])

    request_params = post.call_args[1]['json']['observation']
    assert 'sounds' not in request_params
    mock_upload_sounds.assert_called_once()


def test_create_observation_fail(requests_mock):
    params = {
        'species_guess': 'Pieris rapae',
        # Some invalid data so the observation is rejected...
        'observed_on_string': (datetime.now() + timedelta(days=1)).isoformat(),
        'latitude': 200,
    }

    requests_mock.post(
        f'{API_V0_BASE_URL}/observations.json',
        json=load_sample_data('create_observation_fail.json'),
        status_code=422,
    )

    with pytest.raises(HTTPError) as excinfo:
        create_observation(access_token='valid token', **params)
    assert excinfo.value.response.status_code == 422
    assert 'errors' in excinfo.value.response.json()  # iNat also give details about the errors


def test_update_observation(requests_mock):
    requests_mock.put(
        f'{API_V0_BASE_URL}/observations/17932425.json',
        json=load_sample_data('update_observation_result.json'),
        status_code=200,
    )

    params = {
        'ignore_photos': 1,
        'description': 'updated description v2 !',
    }
    response = update_observation(17932425, access_token='valid token', **params)

    # If all goes well we got a single element representing the updated observation, enclosed in a list.
    assert len(response) == 1
    assert response[0]['id'] == 17932425
    assert response[0]['description'] == 'updated description v2 !'


@patch('pyinaturalist.v0.observations.upload_photos')
@patch('pyinaturalist.v0.observations.put')
def test_update_observation__with_photos(put, mock_upload_photos):
    update_observation(1234, access_token='token', photos='photo.jpg')

    request_params = put.call_args[1]['json']['observation']
    assert 'local_photos' not in request_params
    mock_upload_photos.assert_called_once()


@patch('pyinaturalist.v0.observations.upload_sounds')
@patch('pyinaturalist.v0.observations.put')
def test_update_observation__with_sounds(put, mock_upload_sounds):
    update_observation(1234, access_token='token', sounds='photo.jpg')

    request_params = put.call_args[1]['json']['observation']
    assert 'sounds' not in request_params
    mock_upload_sounds.assert_called_once()


def test_update_nonexistent_observation(requests_mock):
    """When we try to update a non-existent observation, iNat returns an error 410 with
    'observation does not exist'
    """
    requests_mock.put(
        f'{API_V0_BASE_URL}/observations/999999999.json',
        json={'error': 'Cette observation n’existe plus.'},
        status_code=410,
    )

    params = {
        'ignore_photos': 1,
        'description': 'updated description v2 !',
    }

    with pytest.raises(HTTPError) as excinfo:
        update_observation(999999999, access_token='valid token', **params)
    assert excinfo.value.response.status_code == 410
    assert excinfo.value.response.json() == {'error': 'Cette observation n’existe plus.'}


def test_update_observation_not_mine(requests_mock):
    """When we try to update the obs of another user, iNat returns an error 410 with 'obs does not longer exists'."""
    requests_mock.put(
        f'{API_V0_BASE_URL}/observations/16227955.json',
        json={'error': 'Cette observation n’existe plus.'},
        status_code=410,
    )

    params = {
        'ignore_photos': 1,
        'description': 'updated description v2 !',
    }

    with pytest.raises(HTTPError) as excinfo:
        update_observation(16227955, access_token='valid token for another user', **params)
    assert excinfo.value.response.status_code == 410
    assert excinfo.value.response.json() == {'error': 'Cette observation n’existe plus.'}


def test_upload_photos(requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/observation_photos',
        json=load_sample_data('post_observation_photos.json'),
        status_code=200,
    )

    response = upload_photos(1234, BytesIO(), access_token='token')
    assert response[0]['id'] == 1234
    assert response[0]['created_at'] == '2020-09-24T21:06:16.964-05:00'
    assert response[0]['photo']['native_username'] == 'username'


def test_upload_sounds(requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/observation_sounds',
        json=load_sample_data('post_observation_sounds.json'),
        status_code=200,
    )

    response = upload_sounds(233946, BytesIO(), access_token='token')
    assert response[0]['id'] == 233946
    assert response[0]['created_at'] == '2021-05-30T17:36:40.286-05:00'
    assert response[0]['sound']['file_content_type'] == 'audio/mpeg'


def test_delete_observation(requests_mock):
    requests_mock.delete(f'{API_V0_BASE_URL}/observations/24774619.json', status_code=200)
    response = delete_observation(observation_id=24774619, access_token='valid token')
    assert response is None


def test_delete_unexisting_observation(requests_mock):
    """ObservationNotFound is raised if the observation doesn't exists"""
    requests_mock.delete(f'{API_V0_BASE_URL}/observations/24774619.json', status_code=404)

    with pytest.raises(ObservationNotFound):
        delete_observation(observation_id=24774619, access_token='valid token')



from datetime import datetime

from pyinaturalist.constants import TableRow
from pyinaturalist.models import BaseModel, LazyProperty, User, datetime_now_attr, define_model, kwarg


@define_model
class Comment(BaseModel):
    """An observation comment, based on the schema of comments
    from `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    body: str = kwarg
    created_at: datetime = datetime_now_attr
    hidden: bool = kwarg
    id: int = kwarg
    uuid: str = kwarg

    # Lazy-loaded nested model objects
    user: property = LazyProperty(User.from_json)

    # Unused attributes
    # created_at_details: Dict = field(factory=dict)
    # flags: List = field(factory=list)
    # moderator_actions: List = field(factory=list)

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'User': self.user.login,
            'Created at': self.created_at,
            'Comment': self.body,
        }

    def __str__(self) -> str:
        return f'{self.user.login} at {self.created_at}: {self.body}'



from typing import List

from attr import field

from pyinaturalist.constants import TableRow
from pyinaturalist.models import BaseModel, LazyProperty, User, define_model, kwarg


@define_model
class Annotation(BaseModel):
    """An annotation, meaning a **controlled term value** applied by a **user** to an **observation**.
    Based on the schema of annotations from
    `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    concatenated_attr_val: str = kwarg
    controlled_attribute_id: int = kwarg
    controlled_value_id: int = kwarg
    user_id: int = kwarg
    uuid: str = kwarg
    vote_score: int = kwarg
    votes: List = field(factory=list)
    user: property = LazyProperty(User.from_json)

    @property
    def values(self) -> List[str]:
        """Split pipe-delimited annotation values into separate tokens"""
        if not self.concatenated_attr_val:
            return []
        return self.concatenated_attr_val.split('|')

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.controlled_attribute_id,
            'Values': ', '.join(self.values),
            'Votes': self.vote_score,
            'User': self.user.login,
        }

    def __str__(self) -> str:
        return (
            f'[{self.controlled_attribute_id}] {self.concatenated_attr_val} '
            f'({len(self.votes)} votes)'
        )


@define_model
class ControlledTermValue(BaseModel):
    """A controlled term **value**, based on the schema of
    `GET /controlled_terms <https://api.inaturalist.org/v1/docs/#!/Controlled_Terms/get_controlled_terms>`_.
    """

    blocking: bool = kwarg
    id: int = kwarg
    label: str = kwarg
    ontology_uri: str = kwarg
    uri: str = kwarg
    uuid: str = kwarg
    taxon_ids: List[int] = field(factory=list)

    def __str__(self):
        return f'[{self.id}] {self.label}'


@define_model
class ControlledTerm(BaseModel):
    """A controlled term, based on the schema of
    `GET /controlled_terms <https://api.inaturalist.org/v1/docs/#!/Controlled_Terms/get_controlled_terms>`_.
    """

    id: int = kwarg
    is_value: bool = kwarg
    multivalued: bool = kwarg
    label: str = kwarg
    ontology_uri: str = kwarg
    uri: str = kwarg
    uuid: str = kwarg
    taxon_ids: List[int] = field(factory=list)

    values: property = LazyProperty(ControlledTermValue.from_json_list)

    @property
    def value_labels(self) -> str:
        """Combined labels from all controlled term values"""
        return ', '.join([value.label for value in self.values])

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Label': self.label,
            'Values': self.value_labels,
        }

    def __str__(self):
        return f'[{self.id}] {self.label}: {self.value_labels}'



{
    'total_results': 1,
    'page': 1,
    'per_page': 30,
    'results': [
        {
            'photos_locked': False,
            'observations_count': 112516,
            'taxon_schemes_count': 1,
            'ancestry': '48460/1/47120/372739/47158/184884/47201/124417/326777/48740/52747',
            'is_active': True,
            'flag_counts': {'unresolved': 0, 'resolved': 0},
            'wikipedia_url': 'http://en.wikipedia.org/wiki/Polistinae',
            'current_synonymous_taxon_ids': None,
            'iconic_taxon_id': 47158,
            'taxon_photos': [
                {
                    'taxon_id': 343248,
                    'photo': {
                        'flags': [],
                        'type': 'LocalPhoto',
                        'url': 'https://static.inaturalist.org/photos/22765568/square.jpeg?1533696452',
                        'square_url': 'https://static.inaturalist.org/photos/22765568/square.jpeg?1533696452',
                        'native_page_url': 'https://www.inaturalist.org/photos/22765568',
                        'native_photo_id': '22765568',
                        'small_url': 'https://static.inaturalist.org/photos/22765568/small.jpeg?1533696452',
                        'original_url': 'https://static.inaturalist.org/photos/22765568/original.jpeg?1533696452',
                        'attribution': '(c) copper, some rights reserved (CC BY-NC)',
                        'medium_url': 'https://static.inaturalist.org/photos/22765568/medium.jpeg?1533696452',
                        'id': 22765568,
                        'license_code': 'cc-by-nc',
                        'original_dimensions': {'width': 2048, 'height': 1639},
                        'large_url': 'https://static.inaturalist.org/photos/22765568/large.jpeg?1533696452',
                    },
                    'taxon': {
                        'photos_locked': False,
                        'taxon_schemes_count': 1,
                        'ancestry': '48460/1/47120/372739/47158/184884/47201/124417/326777/48740/52747',
                        'min_species_ancestry': '48460,1,47120,372739,47158,184884,47201,124417,326777,48740,52747,343248',
                        'wikipedia_url': 'http://en.wikipedia.org/wiki/Polistinae',
                        'current_synonymous_taxon_ids': None,
                        'iconic_taxon_id': 47158,
                        'created_at': '2014-02-19T09:17:40+00:00',
                        'taxon_changes_count': 0,
                        'complete_species_count': None,
                        'rank': 'subfamily',
                        'extinct': False,
                        'id': 343248,
                        'universal_search_rank': 112516,
                        'ancestor_ids': [
                            48460,
                            1,
                            47120,
                            372739,
                            47158,
                            184884,
                            47201,
                            124417,
                            326777,
                            48740,
                            52747,
                            343248,
                        ],
                        'observations_count': 112516,
                        'is_active': True,
                        'flag_counts': {'unresolved': 0, 'resolved': 0},
                        'min_species_taxon_id': 343248,
                        'rank_level': 27,
                        'atlas_id': None,
                        'parent_id': 52747,
                        'name': 'Polistinae',
                        'default_photo': {
                            'square_url': 'https://static.inaturalist.org/photos/22765568/square.jpeg?1533696452',
                            'attribution': '(c) copper, some rights reserved (CC BY-NC)',
                            'flags': [],
                            'medium_url': 'https://static.inaturalist.org/photos/22765568/medium.jpeg?1533696452',
                            'id': 22765568,
                            'license_code': 'cc-by-nc',
                            'original_dimensions': {'width': 2048, 'height': 1639},
                            'url': 'https://static.inaturalist.org/photos/22765568/square.jpeg?1533696452',
                        },
                        'iconic_taxon_name': 'Insecta',
                        'preferred_common_name': 'Paper Wasps',
                    },
                },
            ],
            'rank_level': 27,
            'taxon_changes_count': 0,
            'atlas_id': None,
            'complete_species_count': None,
            'parent_id': 52747,
            'name': 'Polistinae',
            'rank': 'subfamily',
            'extinct': False,
            'id': 343248,
            'default_photo': {
                'square_url': 'https://static.inaturalist.org/photos/22765568/square.jpeg?1533696452',
                'attribution': '(c) copper, some rights reserved (CC BY-NC)',
                'flags': [],
                'medium_url': 'https://static.inaturalist.org/photos/22765568/medium.jpeg?1533696452',
                'id': 22765568,
                'license_code': 'cc-by-nc',
                'original_dimensions': {'width': 2048, 'height': 1639},
                'url': 'https://static.inaturalist.org/photos/22765568/square.jpeg?1533696452',
            },
            'ancestor_ids': [
                48460,
                1,
                47120,
                372739,
                47158,
                184884,
                47201,
                124417,
                326777,
                48740,
                52747,
            ],
            'iconic_taxon_name': 'Insecta',
            'preferred_common_name': 'Paper Wasps',
            'ancestors': [
                {
                    'observations_count': 32873107,
                    'taxon_schemes_count': 2,
                    'ancestry': '48460',
                    'is_active': True,
                    'flag_counts': {'unresolved': 0, 'resolved': 9},
                    'wikipedia_url': 'http://en.wikipedia.org/wiki/Animal',
                    'current_synonymous_taxon_ids': None,
                    'iconic_taxon_id': 1,
                    'rank_level': 70,
                    'taxon_changes_count': 4,
                    'atlas_id': None,
                    'complete_species_count': None,
                    'parent_id': 48460,
                    'complete_rank': 'order',
                    'name': 'Animalia',
                    'rank': 'kingdom',
                    'extinct': False,
                    'id': 1,
                    'default_photo': {
                        'square_url': 'https://static.inaturalist.org/photos/80678745/square.jpg?1593020613',
                        'attribution': 'Ningún derecho reservado',
                        'flags': [],
                        'medium_url': 'https://static.inaturalist.org/photos/80678745/medium.jpg?1593020613',
                        'id': 80678745,
                        'license_code': 'cc0',
                        'original_dimensions': {'width': 2000, 'height': 2000},
                        'url': 'https://static.inaturalist.org/photos/80678745/square.jpg?1593020613',
                    },
                    'ancestor_ids': [48460, 1],
                    'iconic_taxon_name': 'Animalia',
                    'preferred_common_name': 'Animals',
                },
            ],
            'children': [
                {
                    'observations_count': 88534,
                    'taxon_schemes_count': 0,
                    'ancestry': '48460/1/47120/372739/47158/184884/47201/124417/326777/48740/52747/343248',
                    'is_active': True,
                    'flag_counts': {'unresolved': 0, 'resolved': 0},
                    'wikipedia_url': 'http://en.wikipedia.org/wiki/Polistes',
                    'current_synonymous_taxon_ids': None,
                    'iconic_taxon_id': 47158,
                    'rank_level': 25,
                    'taxon_changes_count': 0,
                    'atlas_id': None,
                    'complete_species_count': None,
                    'parent_id': 343248,
                    'name': 'Polistini',
                    'rank': 'tribe',
                    'extinct': False,
                    'id': 638567,
                    'default_photo': {
                        'square_url': 'https://static.inaturalist.org/photos/2367678/square.jpg?1441941708',
                        'attribution': '(c) Cheryl Harleston López Espino, some rights reserved (CC BY-NC-ND)',
                        'flags': [],
                        'medium_url': 'https://static.inaturalist.org/photos/2367678/medium.jpg?1441941708',
                        'id': 2367678,
                        'license_code': 'cc-by-nc-nd',
                        'original_dimensions': {'width': 900, 'height': 1200},
                        'url': 'https://static.inaturalist.org/photos/2367678/square.jpg?1441941708',
                    },
                    'ancestor_ids': [
                        48460,
                        1,
                        47120,
                        372739,
                        47158,
                        184884,
                        47201,
                        124417,
                        326777,
                        48740,
                        52747,
                        343248,
                        638567,
                    ],
                    'iconic_taxon_name': 'Insecta',
                },
            ],
            'conservation_statuses': [],
            'listed_taxa_count': 0,
            'listed_taxa': [],
            'wikipedia_summary': 'The <b>Polistinae</b> are eusocial wasps closely related to the more familiar yellow jackets, but placed in their own subfamily, containing four tribes; with some 1100 species total, it is the second-most diverse subfamily within the Vespidae, and while most species are tropical or subtropical, they include some of the most frequently encountered large wasps in temperate regions. They are also known as paper wasps, which is a misleading term since other wasps (including yellow jackets) also...',
        }
    ],
}



# flake8: noqa: F405
import pytest

from rich.console import Console
from rich.table import Table

from pyinaturalist.formatters import *
from test.sample_data import *

# Lists of JSON records that can be formatted into tables
TABULAR_RESPONSES = [
    j_comments,
    [j_controlled_term_1, j_controlled_term_2],
    [j_identification_1, j_identification_2],
    [j_observation_1, j_observation_2],
    j_obs_species_counts,
    j_life_list,
    [j_photo_1, j_photo_2_partial],
    [j_place_1, j_place_2],
    j_places_nearby,
    [j_project_1, j_project_2],
    j_search_results,
    [j_taxon_1, j_taxon_2_partial],
    [j_user_1, j_user_2_partial],
]


def get_variations(response_object):
    """Formatting functions should accept any of these variations"""
    return [{'results': [response_object]}, [response_object], response_object]


# TODO: More thorough tests for table content
@pytest.mark.parametrize('response', TABULAR_RESPONSES)
def test_format_table(response):
    table = format_table(response)
    assert isinstance(table, Table)

    def _get_id(value):
        return str(value.get('id') or value.get('record', {}).get('id'))

    # Just make sure at least object IDs show up in the table
    console = Console()
    rendered_table = '\n'.join([str(line) for line in console.render_lines(table)])
    if isinstance(response, list):
        assert all([_get_id(value) in rendered_table for value in response])

    # for obj in response:
    #     assert all([value in rendered_table for value in obj.row.values()])


# TODO: Test content written to stdout. For now, just make sure it doesn't explode.
@pytest.mark.parametrize('response', TABULAR_RESPONSES)
def test_pprint(response):
    pprint(response)


@pytest.mark.parametrize('input', get_variations(j_controlled_term_1))
def test_format_controlled_terms(input):
    assert (
        format_controlled_terms(input)
        == '[12] Plant Phenology: No Evidence of Flowering, Flowering, Fruiting, Flower Budding'
    )


@pytest.mark.parametrize('input', get_variations(j_identification_1))
def test_format_identifications(input):
    expected_str = '[155554373] Species: 60132 (supporting) added on 2021-02-18 20:31:32-06:00 by jkcook'
    assert format_identifications(input) == expected_str


@pytest.mark.parametrize('input', get_variations(j_observation_1))
def test_format_observation(input):
    expected_str = (
        '[16227955] Species: Lixus bardanae observed on 2018-09-05 14:06:00+01:00 '
        'by niconoe at 54 rue des Badauds'
    )
    assert format_observations(input) == expected_str


@pytest.mark.parametrize('input', get_variations(j_project_1))
def test_format_projects(input):
    expected_str = '[8291] PNW Invasive Plant EDDR'
    assert format_projects(input) == expected_str


@pytest.mark.parametrize('input', get_variations(j_place_1))
def test_format_places(input):
    expected_str = '[89191] Conservation Area Riversdale'
    assert format_places(input) == expected_str


def test_format_places__nearby():
    places_str = """
[97394] North America
[97395] Asia
[97393] Oceania
[11770] Mehedinti
[119755] Mahurangi College
[150981] Ceap Breatainn
""".strip()
    assert format_places(j_places_nearby) == places_str


def test_format_search_results():
    expected_str = (
        '[Taxon] [47792] Order: Odonata (Dragonflies and Damselflies)\n'
        '[Place] [113562] Odonates of Peninsular India and Sri Lanka\n'
        '[Project] [9978] Ohio Dragonfly Survey  (Ohio Odonata Survey)\n'
        '[User] [113886] odonatanb (Gilles Belliveau)'
    )
    assert format_search_results(j_search_results) == expected_str


@pytest.mark.parametrize('input', get_variations(j_species_count_1))
def test_format_species_counts(input):
    expected_str = '[48484] Species: Harmonia axyridis (Asian Lady Beetle): 31'
    assert format_species_counts(input) == expected_str


@pytest.mark.parametrize('input', get_variations(j_taxon_1))
def test_format_taxa__with_common_name(input):
    expected_str = '[70118] Species: Nicrophorus vespilloides (Lesser Vespillo Burying Beetle)'
    assert format_taxa(input) == expected_str


@pytest.mark.parametrize('input', get_variations(j_taxon_3_no_common_name))
def test_format_taxon__without_common_name(input):
    assert format_taxa(input) == '[124162] Species: Temnostoma vespiforme'


@pytest.mark.parametrize('input', get_variations(j_user_2_partial))
def test_format_users(input):
    expected_str = '[886482] niconoe (Nicolas Noé)'
    assert format_users(input) == expected_str


def test_simplify_observation():
    simplified_obs = simplify_observations(j_observation_1)
    # Not much worth testing here, just make sure it returns something that can be formatted
    assert format_observations(simplified_obs)



from itertools import groupby
from typing import List

from attr import field

from pyinaturalist.constants import JsonResponse
from pyinaturalist.models import Taxon, TaxonCount, TaxonCounts, define_model, kwarg


@define_model
class LifeListTaxon(TaxonCount):
    """A single taxon in a user's life list"""

    descendant_obs_count: int = field(default=0)
    direct_obs_count: int = field(default=0)

    @property
    def indent_level(self) -> int:
        """Indentation level corresponding to this item's rank level"""
        return int(((70 - self.rank_level) / 5))

    def __str__(self) -> str:
        padding = " " * self.indent_level
        return f'[{self.id:<8}] {padding} {self.rank.title()} {self.name}: {self.count}'


@define_model
class LifeList(TaxonCounts):
    """A user's life list, based on the schema of ``GET /observations/taxonomy``"""

    count_without_taxon: int = field(default=0)
    taxa: List[LifeListTaxon] = field(factory=list, converter=LifeListTaxon.from_json_list)  # type: ignore
    user_id: int = kwarg

    @classmethod
    def from_json(cls, value: JsonResponse, user_id: int = None, **kwargs) -> 'LifeListTaxon':  # type: ignore
        count_without_taxon = value.get('count_without_taxon', 0)
        if 'results' in value:
            value = value['results']

        life_list_json = {'taxa': value, 'user_id': user_id, 'count_without_taxon': count_without_taxon}
        return super(LifeList, cls).from_json(life_list_json)  # type: ignore

    def tree(self):
        """**Experimental**
        Organize this life list into a taxonomic tree

        Returns:
            :py:class:`rich.tree.Tree`
        """
        return make_tree(self.taxa)


def make_tree(taxa: List[Taxon]):
    """Organize a list of taxa into a taxonomic tree. Must contain at least one taxon with
    'Life' (taxon ID 48460) as its parent.

    Returns:
        :py:class:`rich.tree.Tree`
    """

    from rich.tree import Tree

    taxa_by_parent_id = _sort_groupby(taxa, key=lambda x: x.parent_id or -1)

    def make_child_tree(node, taxon):
        """Add a taxon and its children to the specified tree node.
        Base case: leaf taxon (with no children)
        Recursive case: non-leaf taxon (with children)
        """
        node = node.add(taxon.full_name)
        for child in taxa_by_parent_id.get(taxon.id, []):
            node.add(make_child_tree(node, child))
        return node

    tree_root = {'id': 48460, 'name': 'Life', 'rank': 'State of matter'}
    return make_child_tree(Tree('life list', expanded=False), Taxon.from_json(tree_root))


def _sort_groupby(values, key):
    """Apply sorting then grouping using the same key"""
    return {k: list(group) for k, group in groupby(sorted(values, key=key), key=key)}



"""
Functions to access the iNaturalist API v1
See: http://api.inaturalist.org/v1/docs/

Most recent API version tested: 1.3.0
"""
# flake8: noqa: F401, F403
from requests import Response

from pyinaturalist.api_requests import get
from pyinaturalist.constants import API_V1_BASE_URL


def get_v1(endpoint: str, **kwargs) -> Response:
    """Make an API call to iNaturalist.

    Args:
        endpoint: The name of an endpoint resource, not including the base URL e.g. 'observations'
        kwargs: Arguments for :py:func:`.api_requests.request`
    """
    return get(f'{API_V1_BASE_URL}/{endpoint}', **kwargs)


from pyinaturalist.v1.controlled_terms import get_controlled_terms
from pyinaturalist.v1.identifications import get_identifications, get_identifications_by_id
from pyinaturalist.v1.observations import (
    get_observation,
    get_observation_histogram,
    get_observation_identifiers,
    get_observation_observers,
    get_observation_species_counts,
    get_observation_taxonomy,
    get_observations,
)
from pyinaturalist.v1.places import get_places_autocomplete, get_places_by_id, get_places_nearby
from pyinaturalist.v1.projects import get_projects, get_projects_by_id
from pyinaturalist.v1.search import search
from pyinaturalist.v1.taxa import get_taxa, get_taxa_autocomplete, get_taxa_by_id
from pyinaturalist.v1.users import get_user_by_id, get_users_autocomplete



import datetime
from dateutil.tz import tzutc

{
    'total_results': 4,
    'page': 1,
    'per_page': 4,
    'results': [
        {
            'score': 9.062307,
            'type': 'Taxon',
            'matches': ['Odonata', 'Odonati', 'Odonates', 'odonates', 'Odonats'],
            'record': {
                'id': 47792,
                'rank': 'order',
                'rank_level': 40,
                'iconic_taxon_id': 47158,
                'ancestor_ids': [48460, 1, 47120, 372739, 47158, 184884, 47792],
                'is_active': True,
                'min_species_taxon_id': 47792,
                'name': 'Odonata',
                'parent_id': 184884,
                'ancestry': '48460/1/47120/372739/47158/184884',
                'min_species_ancestry': '48460,1,47120,372739,47158,184884,47792',
                'extinct': False,
                'created_at': datetime.datetime(2008, 6, 24, 23, 44, 58, tzinfo=tzutc()),
                'default_photo': {
                    'id': 362106,
                    'license_code': 'cc-by-nc-sa',
                    'attribution': '(c) Vijay Anand Ismavel, some rights reserved (CC BY-NC-SA), uploaded by Dr. Vijay Anand Ismavel MS MCh',
                    'url': 'https://inaturalist-open-data.s3.amazonaws.com/photos/362106/square.jpg?1444643485',
                    'original_dimensions': {'height': 1053, 'width': 1579},
                    'flags': [],
                    'square_url': 'https://inaturalist-open-data.s3.amazonaws.com/photos/362106/square.jpg?1444643485',
                    'medium_url': 'https://inaturalist-open-data.s3.amazonaws.com/photos/362106/medium.jpg?1444643485',
                },
                'taxon_changes_count': 0,
                'taxon_schemes_count': 2,
                'observations_count': 1154269,
                'photos_locked': False,
                'universal_search_rank': 1154269,
                'flag_counts': {'resolved': 0, 'unresolved': 0},
                'current_synonymous_taxon_ids': None,
                'taxon_photos': ['TRUNCATED'],
                'atlas_id': None,
                'complete_species_count': 6300,
                'wikipedia_url': 'http://en.wikipedia.org/wiki/Odonata',
                'complete_rank': 'species',
                'matched_term': 'Odonata',
                'iconic_taxon_name': 'Insecta',
                'preferred_common_name': 'Dragonflies and Damselflies',
            },
        },
        {
            'score': 7.116488,
            'type': 'Place',
            'matches': ['Odonates of Peninsular India and Sri Lanka'],
            'record': {
                'id': 113562,
                'uuid': '41b4c009-438e-47df-a0e2-54b3f05a932f',
                'slug': 'odonates-of-peninsular-india-and-sri-lanka',
                'name': 'Odonates of Peninsular India and Sri Lanka',
                'display_name': 'Odonates of Peninsular India and Sri Lanka',
                'display_name_autocomplete': 'Odonates of Peninsular India and Sri Lanka',
                'place_type': None,
                'admin_level': None,
                'bbox_area': 110.9856964778392,
                'ancestor_place_ids': [97395, 113562],
                'user': {
                    'id': 269948,
                    'login': 'hopeland',
                    'spam': False,
                    'suspended': False,
                    'created_at': '2016-07-09T07:28:15.128Z',
                },
                'geometry_geojson': {'type': 'Polygon', 'coordinates': ['TRUNCATED']},
                'bounding_box_geojson': {'type': 'Polygon', 'coordinates': ['TRUNCATED']},
                'location': [10.1616628185, 78.8619335657],
                'point_geojson': {'type': 'Point', 'coordinates': [78.8619335657, 10.1616628185]},
                'without_check_list': None,
                'observations_count': 130762,
                'universal_search_rank': 130762,
                'names': [],
                'matched_term': 'Odonates of Peninsular India and Sri Lanka',
            },
        },
        {
            'score': 6.9390197,
            'type': 'Project',
            'matches': ['Odonata'],
            'record': {
                'id': 9978,
                'title': 'Ohio Dragonfly Survey  (Ohio Odonata Survey)',
                'title_autocomplete': 'Ohio Dragonfly Survey  (Ohio Odonata Survey)',
                'title_exact': 'Ohio Dragonfly Survey  (Ohio Odonata Survey)',
                'hide_title': False,
                'description': 'This project is to update information on the distribution and abundance of Ohio dragonflies and damselflies. It is an update from the first Ohio Odonata Survey which ran from 1991 to 2000.\r\n\r\nThe first survey was entirely based up museum specimens, new collections of specimens, and published articles in peer reviewed journals. Photography has advanced greatly since then, and this site will utilize digital photos to add to scientifically collected specimens.',
                'slug': 'ohio-dragonfly-survey-ohio-odonata-survey',
                'project_type': 'assessment',
                'banner_color': None,
                'ancestor_place_ids': None,
                'place_ids': None,
                'place_id': 31,
                'user_id': 395663,
                'admins': [{'id': 175160, 'user_id': 370069, 'project_id': 9978, 'role': 'manager'}],
                'rule_place_ids': [31],
                'associated_place_ids': [97394, 1, 31],
                'user_ids': ['TRUNCATED'],
                'location': None,
                'geojson': None,
                'icon': 'https://static.inaturalist.org/projects/9978-icon-span2.jpg?1505527521',
                'icon_file_name': 'JewelAppalach9229.jpg',
                'header_image_url': 'https://static.inaturalist.org/projects/9978-cover.jpg?1521571861',
                'header_image_file_name': '12spot_projectCover.jpg',
                'header_image_contain': False,
                'project_observation_fields': [],
                'terms': '*  Entries must be photos of dragonflies or damselflies taken in Ohio\r\n*  Photos must be identifiable to species. Often 2 or 3 shots show key details such as in field guides extremely helpful.\r\n*  Entries must carry scientific level of data\r\n    -- name of photographer\r\n    -- date photo taken\r\n    -- location as precise as possible.  \r\nMust include:  \r\n    1. County (township if possible)\r\n    2. Site name or description to help pin locaton (can be simple\r\n        like "State Route 56 at Mad River" or "backyard in Athens, Ohio"  or "Big Walnut Creek at confluence with Alum Creek" or something else that will enable to site to located.\r\n   -- GPS coordinates\r\n*  Photos will be vetted and accepted/rejected from the project based upon a) complete data;  b)identifiable photo\r\n',
                'search_parameters': [
                    {'field': 'photos', 'value': True, 'value_bool': True},
                    {
                        'field': 'quality_grade',
                        'value': ['research', 'needs_id'],
                        'value_keyword': ['research', 'needs_id'],
                    },
                    {'field': 'taxon_id', 'value': [47792], 'value_number': [47792]},
                    {'field': 'place_id', 'value': [31], 'value_number': [31]},
                ],
                'search_parameter_fields': {
                    'photos': True,
                    'quality_grade': ['research', 'needs_id'],
                    'taxon_id': [47792],
                    'place_id': [31],
                },
                'subproject_ids': [],
                'project_observation_rules': [
                    {'id': 13151, 'operator': 'wild?', 'operand_type': None, 'operand_id': None},
                    {
                        'id': 14672,
                        'operator': 'georeferenced?',
                        'operand_type': None,
                        'operand_id': None,
                    },
                    {
                        'id': 14673,
                        'operator': 'has_a_photo?',
                        'operand_type': None,
                        'operand_id': None,
                    },
                    {
                        'id': 14674,
                        'operator': 'verifiable?',
                        'operand_type': None,
                        'operand_id': None,
                    },
                    {
                        'id': 14675,
                        'operator': 'in_taxon?',
                        'operand_type': 'Taxon',
                        'operand_id': 47792,
                    },
                    {
                        'id': 14676,
                        'operator': 'observed_in_place?',
                        'operand_type': 'Place',
                        'operand_id': 31,
                    },
                ],
                'rule_preferences': [{'field': 'quality_grade', 'value': 'research,needs_id'}],
                'created_at': datetime.datetime(2016, 11, 30, 19, 10, 5, tzinfo=tzutc()),
                'updated_at': datetime.datetime(2021, 4, 16, 15, 6, 43, tzinfo=tzutc()),
                'last_post_at': datetime.datetime(2021, 5, 3, 21, 23, 56, tzinfo=tzutc()),
                'observations_count': 86452,
                'universal_search_rank': 86898,
                'spam': False,
                'flags': [],
                'site_features': [],
                'umbrella_project_ids': [],
                'prefers_user_trust': False,
                'observation_requirements_updated_at': None,
                'matched_term': 'Ohio Dragonfly Survey  (Ohio Odonata Survey)',
                'is_umbrella': False,
            },
        },
        {
            'score': 4.6454225,
            'type': 'User',
            'matches': ['odonatanb'],
            'record': {
                'id': 113886,
                'login': 'odonatanb',
                'spam': False,
                'suspended': False,
                'created_at': datetime.datetime(2015, 7, 3, 23, 12, 14, tzinfo=tzutc()),
                'login_autocomplete': 'odonatanb',
                'login_exact': 'odonatanb',
                'name': 'Gilles Belliveau',
                'name_autocomplete': 'Gilles Belliveau',
                'orcid': None,
                'icon': 'https://static.inaturalist.org/attachments/users/icons/113886/thumb.jpeg?1475548692',
                'observations_count': 440,
                'identifications_count': 877,
                'journal_posts_count': 0,
                'activity_count': 1317,
                'species_count': 136,
                'universal_search_rank': 440,
                'roles': [],
                'site_id': 5,
                'matched_term': 'odonatanb',
                'icon_url': 'https://static.inaturalist.org/attachments/users/icons/113886/medium.jpeg?1475548692',
            },
        },
    ],
}



import os
from unittest.mock import patch

import pyinaturalist
from pyinaturalist.constants import API_V1_BASE_URL
from pyinaturalist.v1 import get_observation
from test.conftest import load_sample_data


@patch.dict(os.environ, {'GITHUB_REF': 'refs/heads/main'}, clear=True)
def test_user_agent(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_observation.json'),
        status_code=200,
    )
    default_ua = pyinaturalist.DEFAULT_USER_AGENT

    # By default, we have a 'Pyinaturalist' user agent:
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == default_ua

    # But if the user sets a custom one, it is indeed used:
    get_observation(16227955, user_agent='CustomUA')
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'CustomUA'

    # We can also set it globally:
    pyinaturalist.user_agent = 'GlobalUA'
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'GlobalUA'

    # And it persists across requests:
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'GlobalUA'

    # But if we have a global and local one, the local has priority
    get_observation(16227955, user_agent='CustomUA 2')
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'CustomUA 2'

    # We can reset the global settings to the default:
    pyinaturalist.user_agent = pyinaturalist.DEFAULT_USER_AGENT
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == default_ua



from typing import Dict, List, Union

from attr import field

from pyinaturalist.constants import (
    INAT_BASE_URL,
    Coordinates,
    GeoJson,
    JsonResponse,
    ResponseOrResults,
    TableRow,
)
from pyinaturalist.converters import convert_lat_long
from pyinaturalist.models import BaseModel, define_model, kwarg
from pyinaturalist.models.base import ensure_list


def convert_optional_lat_long(obj: Union[Dict, List, None, str]):
    return convert_lat_long(obj) or (0, 0)


@define_model
class Place(BaseModel):
    """A curated or community-contributed place. Handles data from the following endpoints:

    * `GET /places/{id} <https://api.inaturalist.org/v1/docs/#!/Places/get_places_id>`_
    * `GET /places/nearby <https://api.inaturalist.org/v1/docs/#!/Places/get_places_nearby>`_
    * `GET /taxa <https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa>`_  (``establishment_means.place``)
    """

    admin_level: int = kwarg
    ancestor_place_ids: List[str] = field(factory=list)
    bbox_area: float = kwarg
    bounding_box_geojson: GeoJson = field(factory=dict)
    category: str = kwarg  # Either 'standard' or 'community'
    display_name: str = kwarg
    geometry_geojson: GeoJson = field(factory=dict)
    id: int = kwarg
    location: Coordinates = field(converter=convert_optional_lat_long, default=None)
    name: str = kwarg
    place_type: int = kwarg
    slug: str = kwarg

    @classmethod
    def from_json(cls, value: JsonResponse, category: str = None, **kwargs) -> BaseModel:
        value.setdefault('category', category)
        return super(Place, cls).from_json(value)

    @classmethod
    def from_json_list(cls, value: ResponseOrResults, **kwargs) -> List[BaseModel]:
        """Optionally use results from /places/nearby to set Place.category"""
        json_value = dict(ensure_list(value)[0])
        if 'results' in json_value:
            json_value = json_value['results']

        if 'standard' in json_value and 'community' in json_value:
            places = [cls.from_json(item, category='standard') for item in json_value['standard']]
            places += [cls.from_json(item, category='community') for item in json_value['community']]
            return places
        else:
            return super(Place, cls).from_json_list(json_value)

    @property
    def ancestry(self) -> str:
        """Handle slash-delimited 'ancestry' string from ``establishment_means.place``"""
        return '/'.join(self.ancestor_place_ids)

    @ancestry.setter
    def ancestry(self, value: str):
        self.ancestor_place_ids = value.split('/')

    @property
    def url(self) -> str:
        """Info URL on iNaturalist.org"""
        return f'{INAT_BASE_URL}/places/{self.id}'

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Latitude': f'{self.location[0]:9.4f}',
            'Longitude': f'{self.location[1]:9.4f}',
            'Name': self.name,
            'Category': self.category,
            'URL': self.url,
        }

    def __str__(self) -> str:
        return f'[{self.id}] {self.name}'



from datetime import datetime
from dateutil.tz import tzutc

{
    datetime(2020, 1, 1, 14, 0, tzinfo=tzutc()): 1,
    datetime(2020, 1, 1, 15, 0, tzinfo=tzutc()): 13,
    datetime(2020, 1, 1, 16, 0, tzinfo=tzutc()): 0,
    datetime(2020, 1, 1, 17, 0, tzinfo=tzutc()): 0,
    datetime(2020, 1, 1, 18, 0, tzinfo=tzutc()): 0,
    datetime(2020, 1, 1, 19, 0, tzinfo=tzutc()): 3,
    datetime(2020, 1, 1, 20, 0, tzinfo=tzutc()): 2,
    datetime(2020, 1, 1, 21, 0, tzinfo=tzutc()): 1,
    datetime(2020, 1, 1, 22, 0, tzinfo=tzutc()): 1,
    datetime(2020, 1, 1, 23, 0, tzinfo=tzutc()): 0,
    datetime(2020, 1, 2, 0, 0, tzinfo=tzutc()): 1,
    datetime(2020, 1, 2, 1, 0, tzinfo=tzutc()): 0,
    datetime(2020, 1, 2, 2, 0, tzinfo=tzutc()): 0,
    datetime(2020, 1, 2, 3, 0, tzinfo=tzutc()): 1,
}



from datetime import datetime
from dateutil.tz import tzoffset, tzutc

{
    'total_results': 2,
    'page': 1,
    'per_page': 30,
    'results': [
        {
            'quality_grade': 'research',
            'time_observed_at': '2020-08-27T08:57:22+00:00',
            'taxon_geoprivacy': 'open',
            'annotations': [],
            'uuid': 'c44b91f3-534d-48c4-933d-a8b4e1d8f87b',
            'id': 57754375,
            'cached_votes_total': 0,
            'identifications_most_agree': True,
            'species_guess': 'Monarch',
            'identifications_most_disagree': False,
            'tags': [],
            'positional_accuracy': None,
            'comments_count': 2,
            'site_id': 5,
            'license_code': 'cc-by-nc',
            'quality_metrics': [],
            'public_positional_accuracy': None,
            'reviewed_by': [115129, 2852555],
            'oauth_application_id': 3,
            'flags': [],
            'created_at': datetime(2020, 8, 27, 18, 0, 51, tzinfo=tzutc()),
            'description': None,
            'project_ids_with_curator_id': [],
            'updated_at': datetime(2020, 8, 28, 12, 4, 18, tzinfo=tzutc()),
            'sounds': [],
            'place_ids': [
                6712,
                7953,
                9853,
                49224,
                59613,
                64422,
                64423,
                66741,
                78395,
                82256,
                97394,
                145030,
                155145,
                164401,
            ],
            'captive': False,
            'taxon': {
                'is_active': True,
                'ancestry': '48460/1/47120/372739/47158/184884/47157/47224/47922/61244/134169/522900/48663',
                'min_species_ancestry': '48460,1,47120,372739,47158,184884,47157,47224,47922,61244,134169,522900,48663,48662',
                'endemic': False,
                'iconic_taxon_id': 47158,
                'min_species_taxon_id': 48662,
                'threatened': True,
                'rank_level': 10,
                'introduced': False,
                'native': True,
                'parent_id': 48663,
                'name': 'Danaus plexippus',
                'rank': 'species',
                'extinct': False,
                'id': 48662,
                'ancestor_ids': [
                    48460,
                    1,
                    47120,
                    372739,
                    47158,
                    184884,
                    47157,
                    47224,
                    47922,
                    61244,
                    134169,
                    522900,
                    48663,
                    48662,
                ],
                'photos_locked': False,
                'taxon_schemes_count': 6,
                'wikipedia_url': 'http://en.wikipedia.org/wiki/Monarch_butterfly',
                'current_synonymous_taxon_ids': None,
                'created_at': '2008-10-26T20:24:33+00:00',
                'taxon_changes_count': 1,
                'complete_species_count': None,
                'universal_search_rank': 131627,
                'observations_count': 131627,
                'flag_counts': {'unresolved': 1, 'resolved': 7},
                'atlas_id': 1231,
                'default_photo': {
                    'square_url': 'https://static.inaturalist.org/photos/13824507/square.jpg?1545715684',
                    'attribution': '(c) fam-esquivel, some rights reserved (CC BY-NC)',
                    'flags': [],
                    'medium_url': 'https://static.inaturalist.org/photos/13824507/medium.jpg?1545715684',
                    'id': 13824507,
                    'license_code': 'cc-by-nc',
                    'original_dimensions': {'width': 2048, 'height': 1361},
                    'url': 'https://static.inaturalist.org/photos/13824507/square.jpg?1545715684',
                },
                'iconic_taxon_name': 'Insecta',
                'preferred_common_name': 'Monarch',
                'conservation_status': {
                    'user_id': None,
                    'status_name': 'imperiled',
                    'iucn': 40,
                    'authority': 'NatureServe',
                    'geoprivacy': None,
                    'source_id': 8,
                    'place_id': 7953,
                    'status': 's2b',
                },
            },
            'ident_taxon_ids': [
                48460,
                1,
                47120,
                372739,
                47158,
                184884,
                47157,
                47224,
                47922,
                61244,
                134169,
                522900,
                48663,
                48662,
            ],
            'outlinks': [{'source': 'GBIF', 'url': 'http://www.gbif.org/occurrence/2856816457'}],
            'faves_count': 0,
            'ofvs': [],
            'num_identification_agreements': 1,
            'preferences': {'prefers_community_taxon': None},
            'comments': [
                {
                    'moderator_actions': [],
                    'hidden': False,
                    'flags': [],
                    'created_at': '2020-08-28T12:04:18.407Z',
                    'id': 5326888,
                    'created_at_details': {
                        'date': '2020-08-28',
                        'week': 35,
                        'month': 8,
                        'hour': 12,
                        'year': 2020,
                        'day': 28,
                    },
                    'body': 'Thankyou ',
                    'uuid': 'b4d00be4-132d-4e51-bfdd-8515a00f161c',
                    'user': {
                        'created_at': '2020-04-26T15:26:56+00:00',
                        'id': 2852555,
                        'login': 'samroom',
                        'spam': False,
                        'suspended': False,
                        'login_autocomplete': 'samroom',
                        'login_exact': 'samroom',
                        'name': None,
                        'name_autocomplete': None,
                        'orcid': None,
                        'icon': None,
                        'observations_count': 59,
                        'identifications_count': 0,
                        'journal_posts_count': 0,
                        'activity_count': 59,
                        'species_count': 52,
                        'universal_search_rank': 59,
                        'roles': [],
                        'site_id': 5,
                        'icon_url': None,
                    },
                },
                {
                    'moderator_actions': [],
                    'hidden': False,
                    'flags': [],
                    'created_at': '2020-08-28T04:54:19.020Z',
                    'id': 5325706,
                    'created_at_details': {
                        'date': '2020-08-28',
                        'week': 35,
                        'month': 8,
                        'hour': 4,
                        'year': 2020,
                        'day': 28,
                    },
                    'body': 'Neat photo! ',
                    'uuid': '25ac67d5-ba09-47a8-91e3-b6e3590a0b72',
                    'user': {
                        'created_at': '2018-03-18T01:34:25+00:00',
                        'id': 792847,
                        'login': 'ingridt3',
                        'spam': False,
                        'suspended': False,
                        'login_autocomplete': 'ingridt3',
                        'login_exact': 'ingridt3',
                        'name': None,
                        'name_autocomplete': None,
                        'orcid': None,
                        'icon': None,
                        'observations_count': 51,
                        'identifications_count': 29,
                        'journal_posts_count': 0,
                        'activity_count': 80,
                        'species_count': 30,
                        'universal_search_rank': 51,
                        'roles': [],
                        'site_id': 5,
                        'icon_url': None,
                    },
                },
            ],
            'map_scale': None,
            'uri': 'https://www.inaturalist.org/observations/57754375',
            'project_ids': [],
            'community_taxon_id': 48662,
            'geojson': {'coordinates': [-104.71929167, 50.0949055], 'type': 'Point'},
            'owners_identification_from_vision': True,
            'identifications_count': 1,
            'obscured': False,
            'num_identification_disagreements': 0,
            'geoprivacy': None,
            'location': [50.0949055, -104.71929167],
            'votes': [],
            'spam': False,
            'user': {
                'site_id': 5,
                'created_at': '2020-04-26T15:26:56+00:00',
                'id': 2852555,
                'login': 'samroom',
                'spam': False,
                'suspended': False,
                'preferences': {},
                'login_autocomplete': 'samroom',
                'login_exact': 'samroom',
                'name': None,
                'name_autocomplete': None,
                'orcid': None,
                'icon': None,
                'observations_count': 59,
                'identifications_count': 0,
                'journal_posts_count': 0,
                'activity_count': 59,
                'species_count': 52,
                'universal_search_rank': 59,
                'roles': [],
                'icon_url': None,
            },
            'mappable': True,
            'identifications_some_agree': True,
            'project_ids_without_curator_id': [],
            'place_guess': 'Railway Ave, Wilcox, SK, CA',
            'identifications': 'TRUNCATED',
            'project_observations': [],
            'photos': [
                {
                    'id': 92152429,
                    'license_code': 'cc-by-nc',
                    'url': 'https://static.inaturalist.org/photos/92152429/square.jpg?1598551272',
                    'attribution': '(c) samroom, some rights reserved (CC BY-NC)',
                    'original_dimensions': {'width': 1188, 'height': 1188},
                    'flags': [],
                }
            ],
            'observation_photos': [
                {
                    'id': 85887066,
                    'position': 0,
                    'uuid': '6f73a9f2-019f-4566-a393-2c2242a3198d',
                    'photo': {
                        'id': 92152429,
                        'license_code': 'cc-by-nc',
                        'url': 'https://static.inaturalist.org/photos/92152429/square.jpg?1598551272',
                        'attribution': '(c) samroom, some rights reserved (CC BY-NC)',
                        'original_dimensions': {'width': 1188, 'height': 1188},
                        'flags': [],
                    },
                }
            ],
            'faves': [],
            'non_owner_ids': 'TRUNCATED',
            'observed_on': datetime(2020, 8, 27, 8, 57, 22, tzinfo=tzoffset('Etc/UTC', 0)),
        },
        {
            'quality_grade': 'research',
            'time_observed_at': '2020-08-26T22:46:10+00:00',
            'taxon_geoprivacy': 'open',
            'annotations': [],
            'uuid': '588c9ad4-c747-497c-a52f-bf856c30d2ff',
            'id': 57707611,
            'cached_votes_total': 0,
            'identifications_most_agree': True,
            'species_guess': 'Monarch',
            'identifications_most_disagree': False,
            'tags': [],
            'positional_accuracy': 374,
            'comments_count': 0,
            'site_id': 5,
            'license_code': None,
            'quality_metrics': [],
            'public_positional_accuracy': 374,
            'reviewed_by': [35013, 115129, 217571, 412189, 792847, 2173064],
            'oauth_application_id': 3,
            'flags': [],
            'created_at': datetime(2020, 8, 27, 4, 50, 17, tzinfo=tzutc()),
            'description': 'Only one today and likely the last one to eclose from the batch of caterpillars from 10-14 days ago. ',
            'project_ids_with_curator_id': [6905],
            'updated_at': datetime(2020, 11, 2, 19, 26, 40, tzinfo=tzutc()),
            'sounds': [],
            'place_ids': [
                6712,
                7953,
                9853,
                49224,
                59613,
                64422,
                64423,
                66741,
                78395,
                82256,
                97394,
                113642,
                120614,
                120671,
                145030,
                155145,
                164401,
            ],
            'captive': False,
            'taxon': {
                'is_active': True,
                'ancestry': '48460/1/47120/372739/47158/184884/47157/47224/47922/61244/134169/522900/48663',
                'min_species_ancestry': '48460,1,47120,372739,47158,184884,47157,47224,47922,61244,134169,522900,48663,48662',
                'endemic': False,
                'iconic_taxon_id': 47158,
                'min_species_taxon_id': 48662,
                'threatened': True,
                'rank_level': 10,
                'introduced': False,
                'native': True,
                'parent_id': 48663,
                'name': 'Danaus plexippus',
                'rank': 'species',
                'extinct': False,
                'id': 48662,
                'ancestor_ids': [
                    48460,
                    1,
                    47120,
                    372739,
                    47158,
                    184884,
                    47157,
                    47224,
                    47922,
                    61244,
                    134169,
                    522900,
                    48663,
                    48662,
                ],
                'photos_locked': False,
                'taxon_schemes_count': 6,
                'wikipedia_url': 'http://en.wikipedia.org/wiki/Monarch_butterfly',
                'current_synonymous_taxon_ids': None,
                'created_at': '2008-10-26T20:24:33+00:00',
                'taxon_changes_count': 1,
                'complete_species_count': None,
                'universal_search_rank': 131627,
                'observations_count': 131627,
                'flag_counts': {'unresolved': 1, 'resolved': 7},
                'atlas_id': 1231,
                'default_photo': {
                    'square_url': 'https://static.inaturalist.org/photos/13824507/square.jpg?1545715684',
                    'attribution': '(c) fam-esquivel, some rights reserved (CC BY-NC)',
                    'flags': [],
                    'medium_url': 'https://static.inaturalist.org/photos/13824507/medium.jpg?1545715684',
                    'id': 13824507,
                    'license_code': 'cc-by-nc',
                    'original_dimensions': {'width': 2048, 'height': 1361},
                    'url': 'https://static.inaturalist.org/photos/13824507/square.jpg?1545715684',
                },
                'iconic_taxon_name': 'Insecta',
                'preferred_common_name': 'Monarch',
                'conservation_status': {
                    'user_id': None,
                    'status_name': 'imperiled',
                    'iucn': 40,
                    'authority': 'NatureServe',
                    'geoprivacy': None,
                    'source_id': 8,
                    'place_id': 7953,
                    'status': 's2b',
                },
            },
            'ident_taxon_ids': [
                48460,
                1,
                47120,
                372739,
                47158,
                184884,
                47157,
                47224,
                47922,
                61244,
                134169,
                522900,
                48663,
                48662,
            ],
            'outlinks': [],
            'faves_count': 0,
            'ofvs': [],
            'num_identification_agreements': 5,
            'preferences': {'prefers_community_taxon': None},
            'comments': [],
            'map_scale': None,
            'uri': 'https://www.inaturalist.org/observations/57707611',
            'project_ids': [6905],
            'community_taxon_id': 48662,
            'geojson': {'coordinates': [-104.5628221383, 50.433418274], 'type': 'Point'},
            'owners_identification_from_vision': True,
            'identifications_count': 5,
            'obscured': False,
            'num_identification_disagreements': 0,
            'geoprivacy': None,
            'location': [50.433418274, -104.5628221383],
            'votes': [],
            'spam': False,
            'user': {
                'site_id': 5,
                'created_at': '2018-03-18T01:34:25+00:00',
                'id': 792847,
                'login': 'ingridt3',
                'spam': False,
                'suspended': False,
                'preferences': {},
                'login_autocomplete': 'ingridt3',
                'login_exact': 'ingridt3',
                'name': None,
                'name_autocomplete': None,
                'orcid': None,
                'icon': None,
                'observations_count': 51,
                'identifications_count': 29,
                'journal_posts_count': 0,
                'activity_count': 80,
                'species_count': 30,
                'universal_search_rank': 51,
                'roles': [],
                'icon_url': None,
            },
            'mappable': True,
            'identifications_some_agree': True,
            'project_ids_without_curator_id': [],
            'place_guess': 'Michener Dr, Regina, SK, CA',
            'identifications': 'TRUNCATED',
            'project_observations': [
                {
                    'id': 45561247,
                    'uuid': '912138ca-651d-4d5f-99c1-d89e9c023bb5',
                    'project': {'id': 6905},
                    'user_id': 35013,
                    'preferences': {'allows_curator_coordinate_access': True},
                    'user': {
                        'id': 35013,
                        'login': 'a_b',
                        'spam': False,
                        'suspended': False,
                        'created_at': '2014-04-09T03:27:04+00:00',
                        'login_autocomplete': 'a_b',
                        'login_exact': 'a_b',
                        'name': 'Andrea Benville',
                        'name_autocomplete': 'Andrea Benville',
                        'orcid': None,
                        'icon': 'https://static.inaturalist.org/attachments/users/icons/35013/thumb.jpg?1505750038',
                        'observations_count': 787,
                        'identifications_count': 533,
                        'journal_posts_count': 0,
                        'activity_count': 1320,
                        'species_count': 432,
                        'universal_search_rank': 787,
                        'roles': [],
                        'site_id': 5,
                        'icon_url': 'https://static.inaturalist.org/attachments/users/icons/35013/medium.jpg?1505750038',
                    },
                }
            ],
            'photos': [
                {
                    'id': 92075215,
                    'license_code': None,
                    'url': 'https://static.inaturalist.org/photos/92075215/square.jpg?1598503820',
                    'attribution': '(c) ingridt3, all rights reserved',
                    'original_dimensions': {'width': 1183, 'height': 842},
                    'flags': [],
                }
            ],
            'observation_photos': [
                {
                    'id': 85814399,
                    'position': 0,
                    'uuid': '98dd67ae-79b4-49c6-934b-6a814ea6d6ab',
                    'photo': {
                        'id': 92075215,
                        'license_code': None,
                        'url': 'https://static.inaturalist.org/photos/92075215/square.jpg?1598503820',
                        'attribution': '(c) ingridt3, all rights reserved',
                        'original_dimensions': {'width': 1183, 'height': 842},
                        'flags': [],
                    },
                }
            ],
            'faves': [],
            'non_owner_ids': 'TRUNCATED',
            'observed_on': datetime(2020, 8, 26, 22, 46, 10, tzinfo=tzoffset('Etc/UTC', 0)),
        },
    ],
}



from datetime import datetime
from typing import Dict, List, Optional

from attr import field

from pyinaturalist.constants import INAT_BASE_URL, Coordinates, JsonResponse, TableRow
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    User,
    coordinate_pair,
    datetime_attr,
    datetime_now_attr,
    define_model,
    kwarg,
)
from pyinaturalist.models.observation_field import ObservationField


@define_model
class ProjectObservation(ObservationField):
    """Metadata about an observation that has been added to a project"""

    id: int = kwarg
    preferences: Dict = field(factory=dict)  # Example: {'allows_curator_coordinate_access': True}
    project: Dict = field(factory=dict)  # Example: {'id': 24237}
    user_id: int = kwarg
    uuid: str = kwarg
    user: property = LazyProperty(User.from_json)


@define_model
class ProjectObservationField(ObservationField):
    """An :py:class:`.ObservationField` with additional project-specific information"""

    project_observation_field_id: int = kwarg
    position: int = kwarg
    required: bool = kwarg

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'ProjectObservationField':
        """Flatten out nested values"""
        obs_field = value['observation_field']
        obs_field['project_observation_field_id'] = value['id']
        obs_field['position'] = value['position']
        obs_field['required'] = value['required']
        return super(ProjectObservationField, cls).from_json(obs_field, **kwargs)  # type: ignore


@define_model
class ProjectUser(User):
    """A :py:class:`.User` with additional project-specific information"""

    project_id: int = kwarg
    project_user_id: int = kwarg
    role: str = kwarg

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'ProjectUser':
        """Flatten out nested values"""
        user = value['user']
        user['project_id'] = value['project_id']
        user['project_user_id'] = value['id']
        user['role'] = value['role']
        return super(ProjectUser, cls).from_json(user, **kwargs)  # type: ignore


@define_model
class Project(BaseModel):
    """A dataclass containing information about a project, matching the schema of
    `GET /projects <https://api.inaturalist.org/v1/docs/#!/Projects/get_projects>`_.
    """

    banner_color: str = kwarg
    created_at: datetime = datetime_now_attr
    description: str = kwarg
    header_image_contain: bool = kwarg
    header_image_file_name: str = kwarg
    header_image_url: str = kwarg
    hide_title: bool = kwarg
    icon_file_name: str = kwarg
    icon: str = kwarg
    id: int = kwarg
    is_umbrella: bool = kwarg
    location: Optional[Coordinates] = coordinate_pair
    place_id: int = kwarg
    prefers_user_trust: bool = kwarg
    project_type: str = kwarg
    slug: str = kwarg
    terms: str = kwarg
    title: str = kwarg
    updated_at: Optional[datetime] = datetime_attr
    user_id: int = kwarg

    # Nested collections
    project_observation_rules: List[Dict] = field(factory=list)
    rule_preferences: List[Dict] = field(factory=list)
    search_parameters: List[Dict] = field(factory=list)
    site_features: List[Dict] = field(factory=list)
    user_ids: List[int] = field(factory=list)

    # Lazy-loaded nested model objects
    admins: property = LazyProperty(ProjectUser.from_json_list)
    project_observation_fields: property = LazyProperty(ProjectObservationField.from_json_list)
    user: property = LazyProperty(User.from_json)

    # Unused attributes
    # flags: List = field(factory=list)
    # latitude: float = kwarg
    # longitude: float = kwarg

    # Aliases
    @property
    def obs_fields(self):
        return self.project_observation_fields

    @property
    def obs_rules(self):
        return self.project_observation_rules

    @property
    def url(self) -> str:
        """Info URL on iNaturalist.org"""
        return f'{INAT_BASE_URL}/projects/{self.id}'

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Title': self.title,
            'Type': self.project_type,
            'URL': self.url,
        }

    def __str__(self) -> str:
        return f'[{self.id}] {self.title}'
